# -*- coding: utf-8 -*-
# autostart = false


import re
from os import environ
from subprocess import run
from keyszer.config_api import *
from keyszer.lib.logger import *

# Keyszer-specific config settings - REMOVE OR SET TO DEFAULTS FOR DISTRIBUTION
dump_diagnostics_key(Key.F15)   # default key: F15
emergency_eject_key(Key.F16)    # default key: F16
timeouts(
    multipurpose    = 1,        # default: 1 sec
    suspend         = 1,        # default: 1 sec, try 0.1 sec for touchpads
)

# # THESE NEW API FUNCTIONS ARE NOT ACTIVE YET. 
# # THEY WON'T WORK UNTIL KEYSZER IS UPDATED TO SUPPORT THEM.
# #
# throttle_delays(
#     key_pre_delay_ms    = 0,    # default: 0 ms, range: 0 to 150 ms, suggested: 1-50 ms
#     key_post_delay_ms   = 0,    # default: 0 ms, range: 0 to 150 ms, suggested: 1-100 ms
# )
# environment(
#     distro_name     = "Ubuntu",
#     session_type    = "Wayland",
#     desktop_env     = "GNOME",
#     shell_ext       = "xremap",
# )


#############################  ENVIRONMENT  ##############################
###                                                                    ###
###                                                                    ###
###      ███████ ███    ██ ██    ██ ██ ██████   ██████  ███    ██      ###
###      ██      ████   ██ ██    ██ ██ ██   ██ ██    ██ ████   ██      ###
###      █████   ██ ██  ██ ██    ██ ██ ██████  ██    ██ ██ ██  ██      ###
###      ██      ██  ██ ██  ██  ██  ██ ██   ██ ██    ██ ██  ██ ██      ###
###      ███████ ██   ████   ████   ██ ██   ██  ██████  ██   ████      ###
###                                                                    ###
###                                                                    ###
##########################################################################
# Set up some useful environment variables

debug(f'\t####################   Welcome to Toshy!   ####################\n', ctx="!!")
debug(f'\tThis is a highly customized fork of the config file that powers\
        \n\tKinto.sh, by Ben Reaves (https://kinto.sh)\n', ctx="##")
debug(f'\tAll credit for the basis of this goes to Ben Reaves\
        \n\t(https://github.com/rbreaves/) and much assistance\
        \n\twas provided by Josh Goebel, the developer of the\
        \n\txkeysnail fork "keyszer" (http://github.com/joshgoebel/keyszer)\n', ctx="##")

DISTRO_NAME = None
DESKTOP_ENV = None
distro_name = ""
desktop_env = ""

with open('/etc/os-release', 'r') as f:
    for line in f:
        if line.startswith('NAME='):
            distro_name = line.split('=')[1].strip().strip('"')
            break
if not distro_name:
    debug(f"Distro name couldn't be found in /etc/os-release!")

distro_names = {                # Only need to list distros that need special overrides
    'elementary':               'eos',
    'Fedora':                   'fedora',
    'Manjaro':                  'manjaro',
    'KDE Neon':                 'neon',
    'Pop!_OS':                  'popos',
    'Ubuntu':                   'ubuntu',
}
for k, v in distro_names.items():
    # debug(f'{k = :<10} {v = :<10}')
    if re.search(k, distro_name, re.I):
        DISTRO_NAME = v
        break

desktop_env = environ.get("XDG_SESSION_DESKTOP") or environ.get("XDG_CURRENT_DESKTOP")

width = 14
if not desktop_env:
    desktop_env = None
    debug("ERROR: Desktop Environment not found in XDG_SESSION_DESKTOP or XDG_CURRENT_DESKTOP.")
    debug("ERROR: Config file will not be able to adapt automatically to Desktop Environment.")

# Produce a simplified desktop environment name
de_names = {
    'Budgie':                   'budgie',
    'Cinnamon':                 'cinnamon',
    'Deepin':                   'deepin',
    'Enlightenment':            'enlightenment',
    'GNOME':                    'gnome',
    'Hyprland':                 'hypr',
    'IceWM':                    'icewm',
    'KDE':                      'kde',
    'LXDE':                     'lxde',
    'LXQt':                     'lxqt',
    'MATE':                     'mate',
    'Pantheon':                 'pantheon',
    'Plasma':                   'kde',
    'SwayWM':                   'sway',
    'Ubuntu':                   'gnome',    # "Ubuntu" in XDG_CURRENT_DESKTOP, but DE is GNOME
    'Unity':                    'unity',
    'Xfce':                     'xfce',
}
for k, v in de_names.items():
    # debug(f'{k = :<10} {v = :<10}')
    if re.search(k, desktop_env, re.I):
        DESKTOP_ENV = v
    # if DESKTOP_ENV == 'kde' and re.search('neon', DISTRO_NAME, re.I):
    #     DESKTOP_ENV = 'kde_neon'         # override 'kde' for KDE Neon
    #     break
    if DESKTOP_ENV:
        break
if not DESKTOP_ENV:
    debug(f'Desktop Environment not in de_names list! Should fix this.\n\t{desktop_env = }', ctx="EE")

debug(f'User config sees: {DISTRO_NAME = }', ctx="CG")
debug(f'User config sees: {DESKTOP_ENV = }', ctx="CG")



#################  VARIABLES  ####################
###                                            ###
###                                            ###
###      ██    ██  █████  ██████  ███████      ###
###      ██    ██ ██   ██ ██   ██ ██           ###
###      ██    ██ ███████ ██████  ███████      ###
###       ██  ██  ██   ██ ██   ██      ██      ###
###        ████   ██   ██ ██   ██ ███████      ###
###                                            ###
###                                            ###
##################################################
# Establish important variables here

# Short names for the string and Unicode processing helper functions
ST = to_US_keystrokes           # was to_keystrokes originally
UC = unicode_keystrokes

# Variable to enable or disable the "Mac Numpad" feature by default:
# This forces the numeric keypad to always be a numeric keypad, 
# regardless of the state of NumLock. And the NumLock key becomes "Clear" (Escape)
# Turn NumLock off in this state with Option-NumLock (or Fn-NumLock on an Apple keyboard)
mac_numpad = True               # Default: True

# Variable to enable media_arrows_fix
# Makes arrow keys with media (Play_Pause/Stop/Rew/Fwd) functions 
# become PgUp/PgDn/Home/End keys when used with Fn key, like a MacBook
media_arrows_fix = True         # Default: False

# Set to True to enable modmapping CapsLock to Cmd (RIGHT_CTRL) for your kbtype
Caps2Cmd = False

# Set this variable to True to enable Sublime keybindings in VSCodes
VSCode_Sublime = False

# Trick to negate the need to put quotes around all the key labels 
# inside the "lists of dicts" to be given to matchProps() function.
# Makes the variables evaluate to equivalent strings inside the dicts. 
# Provides for nice syntax highlighting, visual separation of key:value. 
cls = 'cls'                     # key label for matchProps() arg to match: wm_class
nme = 'nme'                     # key label for matchProps() arg to match: wm_name
dvn = 'dvn'                     # key label for matchProps() arg to match: device_name
not_cls = 'not_cls'             # key label for matchProps() arg to NEGATIVE match: wm_class
not_nme = 'not_nme'             # key label for matchProps() arg to NEGATIVE match: wm_name
not_dvn = 'not_dvn'             # key label for matchProps() arg to NEGATIVE match: device_name
nlk = 'nlk'                     # key label for matchProps() arg to match: numlock_on
clk = 'clk'                     # key label for matchProps() arg to match: capslock_on
cse = 'cse'                     # key label for matchProps() arg to enable: case sensitivity
lst = 'lst'                     # key label for matchProps() arg to pass in a [list] of {dicts}
dbg = 'dbg'                     # key label for matchProps() arg to set debugging info string

# global variables for the isDoubleTap() function
tapTime1 = time.time()
tapInterval = 0.24
tapCount = 0
last_dt_combo = None



###################################  CUSTOM FUNCTIONS  ####################################
###                                                                                     ###
###                                                                                     ###
###      ███████ ██    ██ ███    ██  ██████ ████████ ██  ██████  ███    ██ ███████      ###
###      ██      ██    ██ ████   ██ ██         ██    ██ ██    ██ ████   ██ ██           ###
###      █████   ██    ██ ██ ██  ██ ██         ██    ██ ██    ██ ██ ██  ██ ███████      ###
###      ██      ██    ██ ██  ██ ██ ██         ██    ██ ██    ██ ██  ██ ██      ██      ###
###      ██       ██████  ██   ████  ██████    ██    ██  ██████  ██   ████ ███████      ###
###                                                                                     ###
###                                                                                     ###
###########################################################################################


def isKBtype(kbtype, dbg=None):
    """Check the keyboard type for conditional modmaps and keymaps"""
    def _isKBtype(ctx):
        logging_enabled = False
        dvn = ctx.device_name
        if logging_enabled:
            print(f'## ###  Debugging isKBtype(): {dbg = }')
            print(f'\tDevice name: {dvn = }')
            print(f'\tLooking for keyboard type: {kbtype = }')
            print(f'\tKeyboard list to be searched: {kbtype_lists[kbtype]}')
            print(f'\tDevice name in list: {[dvn in kbtype_lists[kbtype]]}')
            print(f'\tIs there a regex match: {re.search(kbtype, dvn, re.I)}')

        return True if ( dvn in kbtype_lists[kbtype] or re.search(kbtype, dvn, re.I) ) else False

    return _isKBtype


def isDoubleTap(dt_combo):
    """
    Simplistic detection of double-tap of a key. Blocks single-tap function.
    Only cares about the 'real' key in a combo of Mods+key
    """
    def _isDoubleTap():
        global tapTime1
        global tapInterval
        global tapCount
        global last_dt_combo
        _tapTime = time.time()
        # This first "if" block has a logic defect, if a different key in the
        # same keymap is also set up to send the same "dt_combo" value.
        if tapCount == 1 and last_dt_combo != dt_combo:
            debug(f'## isDoubleTap: \n\tDifferent combo: \n\t{last_dt_combo, dt_combo=}')
            last_dt_combo = None
            tapCount = 0
        # 2nd tap beyond time interval? Treat as new double-tap cycle.
        if tapCount == 1 and _tapTime - tapTime1 >= tapInterval:
            debug(f'## isDoubleTap: \n\tTime diff (too long): \n\t{_tapTime - tapTime1=}')
            tapCount = 0
        # Try to keep held key from producing repeats of dt_combo.
        # If repeat rate very slow or delay very short, this won't work well. 
        if tapCount == 1 and _tapTime - tapTime1 < 0.07:
            debug(f'## isDoubleTap: \n\tTime diff (too short): \n\t{_tapTime - tapTime1=}')
            tapCount = 0
            return None
        # 2nd tap within interval window? Reset cycle & send dt_combo.
        if tapCount == 1 and _tapTime - tapTime1 < tapInterval:
            debug(f'## isDoubleTap: \n\tTime diff (just right): \n\t{_tapTime - tapTime1=}')
            tapCount = 0
            tapTime1 = 0.0
            return dt_combo
        # New cycle? Set count = 1, tapTime1 = now. Send nothing. 
        if tapCount == 0:
            debug(f'## isDoubleTap: \n\tTime diff (1st cycle): \n\t{_tapTime - tapTime1=}')
            last_dt_combo = dt_combo
            tapCount = 1
            tapTime1 = _tapTime
            return None
    return _isDoubleTap


def toggle_media_arrows_fix():
    """Toggle the value of the media_arrows_fix variable"""
    # Needs "from subprocess import run" somewhere
    def _toggle_media_arrows_fix():
        global media_arrows_fix
        media_arrows_fix = not media_arrows_fix
        if media_arrows_fix:
            run('notify-send -u critical ALERT "Toshy Media Arrows Fix is now ENABLED.\
                \rMedia function arrow keys will be PgUp/PgDn/Home/End\
                \rwhen used with the Fn key.\
                \rDisable with Shift+Opt+Cmd+M."', shell=True)
            debug("Media Arrows Fix is now ENABLED.")
        else:
            run('notify-send -u critical ALERT "Toshy Media Arrows Fix is now DISABLED.\
                \rRe-enable with Shift+Opt+Cmd+M."', shell=True)
            debug("Media Arrows Fix is now DISABLED.")

    return _toggle_media_arrows_fix


def matchProps(
    dpm=None,                                   # None    (dummy parameter)
    cls=None, nme=None, dvn=None,               # string  (positive match)
    not_cls=None, not_nme=None, not_dvn=None,   # string  (negative match)
    nlk=None, clk=None, cse=None,               # bool
    lst=None,                                   # list of dicts (positive)
    not_lst=None,                               # list of dicts (negative)
    dbg=None,                                   # string  (debugging info)
):
    """
    ### Match all given properties to current window context.   \n
    - Parameters must be _named_, no positional arguments.      \n
    - All parameters optional, but at least one must be given.  \n
    - Defaults to case insensitive matching of:                 \n
        - WM_CLASS, WM_NAME, device_name                        \n
    - To negate/invert regex pattern match use:                 \n
        - `not_cls` `not_nme` `not_dvn` params or...            \n
        - "^(?:(?!^pattern$).)*$"                               \n
    - To force case insensitive pattern match use:              \n 
        - "^(?i:pattern)$" or...                                \n
        - "^(?i)pattern$"                                       \n

    ### Accepted Parameters:                                    \n
    `dpm` = IGNORE THIS - For enforcing use of named parameters \n
    `cls` = WM_CLASS    (regex/string) [xprop WM_CLASS]         \n
    `nme` = WM_NAME     (regex/string) [xprop _NET_WM_NAME]     \n
    `dvn` = Device Name (regex/string) [keyszer --list-devices] \n
    `not_cls` = `cls` but inverted, matches when "not"          \n
    `not_nme` = `nme` but inverted, matches when "not"          \n
    `not_dvn` = `dvn` but inverted, matches when "not"          \n
    `nlk` = Num Lock LED state         (bool)                   \n
    `clk` = Caps Lock LED state        (bool)                   \n
    `cse` = Case Sensitive matching    (bool)                   \n
    `lst` = List of dicts of the above arguments                \n
    `not_lst` = `lst` but inverted, matches when "not"          \n
    `dbg` = Debugging info             (string)                 \n

    ### Negative match parameters: 
    - `not_cls`|`not_nme`|`not_dvn`                             \n
    Parameters take the same regex patterns as `cls`|`nme`|`dvn`\n
    but result in a True condition only if pattern is NOT found.\n
    Negative parameters cannot be used together with the normal \n
    positive matching equivalent parameter in same instance.    \n

    ### List of Dicts parameter: `lst`|`not_lst`
    A [list] of {dicts} with each dict containing 1 to 6 of the \n
    named parameters above, to be processed recursively as args.\n

    ### Debugging info parameter: `dbg`
    A string that will print as part of logging output. Use to  \n
    help identify origin of logging output.                     \n
    -                                                           \n
    """
    # Reference for successful negative lookahead pattern, and explanation of why it works:
    # https://stackoverflow.com/questions/406230/regular-expression-to-match-a-line-that-doesnt-contain-a-word

    logging_enabled = False

    allowed_params  = (cls, nme, dvn, not_cls, not_nme, not_dvn, nlk, clk, cse, lst, not_lst, dbg)
    lst_dct_params  = (cls, nme, dvn, not_cls, not_nme, not_dvn, nlk, clk, cse)
    string_params   = (cls, nme, dvn, not_cls, not_nme, not_dvn, dbg)

    dct_param_strs  = [
        'cls', 'nme', 'dvn', 'not_cls', 'not_nme', 'not_dvn', 'nlk', 'clk', 
        'cse', 'lst', 'not_lst', 'dbg'
    ]

    if dpm is not None: raise ValueError(f"\n\n(EE) matchProps requires named parameters\n")
    if all([x is None for x in allowed_params]): 
        raise ValueError(f"\n\n(EE) matchProps received no valid argument\n")
    if any([x not in (True, False, None) for x in (nlk, clk, cse)]): 
        raise TypeError(f"\n\n(EE) matchProps params 'nlk'|'clk'|'cse' are bools\n")
    if any([x is not None and not isinstance(x, str) for x in string_params]): raise TypeError(\
        f"\n\n(EE) matchProps params 'cls|nme|dvn|not_cls|not_nme|not_dvn|dbg' must be strings\n")
    if cls and not_cls or nme and not_nme or dvn and not_dvn or lst and not_lst: raise ValueError(\
        f"\n\n(EE) matchProps - don't mix positive and negative match params for same property\n")

    # consolidate positive and negative matching params into new vars
    _lst = not_lst if lst is None else lst
    _cls = not_cls if cls is None else cls
    _nme = not_nme if nme is None else nme
    _dvn = not_dvn if dvn is None else dvn

    # process lists of conditions
    if _lst is not None:
        if any([x is not None for x in lst_dct_params]): 
            raise TypeError(f"\n\n(EE) matchProps param 'lst|not_lst' must be used alone\n")
        if not isinstance(_lst, list) or not all(isinstance(item, dict) for item in _lst): 
            raise TypeError(f"\n\n(EE) matchProps param 'lst|not_lst' wants a [list] of {{dicts}}\n")
        # verify that every dict only contains valid parameter names
        for dct in _lst:
            for prm in list(dct.keys()):
                if prm not in dct_param_strs:
                    print(f"\n(EE) Invalid matchProps() parameter: {prm}\n")
                    print(f"(DD) Parameter is in dict: {dct}\n")
                    print(f"(DD) Dict is in list:")
                    for x in _lst:
                        print(f"(DD)   {x}")
                    raise ValueError(f"\n\n(EE) matchProps - invalid parameter found in dict in list\n")

        def _matchProps_Lst(ctx):
            if not_lst is not None:
                if logging_enabled: print(f"## _matchProps_Lst()[not_lst] ## {dbg=}")
                return ( None if next(filter(lambda dct: 
                                    matchProps(**dct)(ctx), not_lst), False) else True )
            else:
                if logging_enabled: print(f"## _matchProps_Lst()[lst] ## {dbg=}")
                return next(filter(lambda dct: matchProps(**dct)(ctx), lst), False)

        return _matchProps_Lst

    # compile case insentitive regex object for given params unless cse=True
    if _cls is not None: cls_rgx = re.compile(_cls) if cse else re.compile(_cls, re.I)
    if _nme is not None: nme_rgx = re.compile(_nme) if cse else re.compile(_nme, re.I)
    if _dvn is not None: dvn_rgx = re.compile(_dvn) if cse else re.compile(_dvn, re.I)

    def _matchProps(ctx):
        cnd_lst = []
        if _cls is not None:                            # param is given?
            if not_cls is not None:                     # negative match requested?
                cnd_lst.append( None if re.search(cls_rgx, ctx.wm_class) else True )
            else:                                       # positive match requested?
                cnd_lst.append( re.search(cls_rgx, ctx.wm_class) )
        if _nme is not None:                            # param is given?
            if not_nme is not None:                     # negative match requested?
                cnd_lst.append( None if re.search(nme_rgx, ctx.wm_name) else True )
            else:                                       # positive match requested?
                cnd_lst.append( re.search(nme_rgx, ctx.wm_name) )
        if _dvn is not None:                            # param is given?
            if not_dvn is not None:                     # negative match requested?
                cnd_lst.append( None if re.search(dvn_rgx, ctx.device_name) else True )
            else:                                       # positive match requested?
                cnd_lst.append( re.search(dvn_rgx, ctx.device_name) )

        # these two MUST check explicitly for "not None" because external input is bool
        if nlk is not None: cnd_lst.append( nlk is ctx.numlock_on  )
        if clk is not None: cnd_lst.append( clk is ctx.capslock_on )

        if logging_enabled: # and all(cnd_lst): # << to show only "True" condition lists
            print(f'####  CND_LST ({all(cnd_lst)})  ####  {dbg=}')
            for elem in cnd_lst:
                print('##', re.sub('^.*span=.*\), ', '', str(elem)).replace('>',''))
            print('-------------------------------------------------------------------')

        return all(cnd_lst)

    return _matchProps


def toRgxStr(lst_of_str):
    """
    Convert a list of strings into a casefolded             \n
    concatenated regex pattern string.
    """
    if not isinstance(lst_of_str, list): raise TypeError(\
        f"\n\n###  toRgxStr wants a list of strings  ###\n")
    elif isinstance(lst_of_str, list):
        if any([not isinstance(x, str) for x in lst_of_str]): raise TypeError(\
            f"\n\n###  toRgxStr wants a list of strings  ###\n")
        lst_of_str_clean = [str(x).replace('^','').replace('$','') for x in lst_of_str]
        return "|".join(str('^'+x.casefold()+'$') for x in lst_of_str_clean)


def negRgx(rgx_str):
    """
    Invert/convert a positive match regex pattern into      \n
    a negative lookahead regex pattern.
    """
    # remove any ^$
    rgx_str_strip = str(rgx_str).replace('^','').replace('$','')
    # add back ^$, but only around ENTIRE STRING (ignore any vertical bars/pipes)
    rgx_str_add = str('^'+rgx_str_strip+'$')
    # convert ^$ to complicated negative lookahead pattern that actually works
    neg_rgx_str = str(rgx_str_add).replace('^','^(?:(?!^').replace('$','$).)*$')
    return neg_rgx_str



######################  LISTS  #######################
###                                                ###
###                                                ###
###      ██      ██ ███████ ████████ ███████       ###
###      ██      ██ ██         ██    ██            ###
###      ██      ██ ███████    ██    ███████       ###
###      ██      ██      ██    ██         ██       ###
###      ███████ ██ ███████    ██    ███████       ###
###                                                ###
###                                                ###
######################################################

# Use the following for testing terminal keymaps
# terminals = [ "", ... ]
# xbindkeys -mk
terminals = [
    "alacritty",
    "cutefish-terminal",
    "deepin-terminal",
    "eterm",
    "gnome-terminal",
    "guake",
    "hyper",
    "io.elementary.terminal",
    "kinto-gui.py",
    "kitty",
    "Kgx",                      # GNOME Console terminal app
    "konsole",
    "lxterminal",
    "mate-terminal",
    "org.gnome.Console",
    "qterminal",
    "st",
    "sakura",
    "station",
    "tabby",
    "terminator",
    "termite",
    "tilda",
    "tilix",
    "urxvt",
    "xfce4-terminal",
    "xterm",
]
terminals = [x.casefold() for x in terminals]
termStr = toRgxStr(terminals)

terminals_lod = [
    {cls:"^alacritty$"                  },
    {cls:"^cutefish-terminal$"          },
    {cls:"^deepin-terminal$"            },
    {cls:"^eterm$"                      },
    {cls:"^gnome-terminal$"             },
    {cls:"^gnome-terminal-server$"      },
    {cls:"^guake$"                      },
    {cls:"^hyper$"                      },
    {cls:"^io.elementary.terminal$"     },
    {cls:"^kinto-gui.py$"               },
    {cls:"^kitty$"                      },
    {cls:"^Kgx$"                        },
    {cls:"^konsole$"                    },
    {cls:"^lxterminal$"                 },
    {cls:"^mate-terminal$"              },
    {cls:"^org.gnome.Console$"          },
    {cls:"^qterminal$"                  },
    {cls:"^st$"                         },
    {cls:"^sakura$"                     },
    {cls:"^station$"                    },
    {cls:"^tabby$"                      },
    {cls:"^terminator$"                 },
    {cls:"^termite$"                    },
    {cls:"^tilda$"                      },
    {cls:"^tilix$"                      },
    {cls:"^urxvt$"                      },
    {cls:"^xfce4-terminal$"             },
    {cls:"^xterm$"                      },
]

mscodes = [
    "code",
    "vscodium",
    "code - oss",
]
mscodes = [x.casefold() for x in mscodes]
mscodeStr = toRgxStr(mscodes)

mscodes_lod = [
    {cls:"^code$"},
    {cls:"^vscodium$"},
    {cls:"^code - oss$"},
]

sublimes = [
    "sublime_text",
    "subl",
]
sublimes = [x.casefold() for x in sublimes]
sublimeStr = toRgxStr(sublimes)

# Add remote desktop clients & VM software here
# Ideally we'd only exclude the client window,
# but that may not be easily done.
remotes = [
    "Gnome-boxes",
    "org.remmina.Remmina",
    "Nxplayer.bin",
    "remmina",
    "qemu-system-.*",
    "qemu",
    "Spicy",
    "Virt-manager",
    "VirtualBox",
    "VirtualBox Machine",
    "xfreerdp",
    "Wfica",
]
remotes = [x.casefold() for x in remotes]
remoteStr = toRgxStr(remotes)

remotes_lod = [
    {cls:"^Gnome-boxes$"                },
    {cls:"^org.remmina.Remmina$",       not_nme:"^Remmina Remote Desktop Client$|^Remote Connection Profile$"},
    {cls:"^Nxplayer.bin$"               },
    {cls:"^remmina$"                    },
    {cls:"^qemu-system-.*$"             },
    {cls:"^qemu$"                       },
    {cls:"^Spicy$"                      },
    {cls:"^Virt-manager$"               },
    {cls:"^VirtualBox$"                 },
    {cls:"^VirtualBox Machine$"         },
    {cls:"^xfreerdp$"                   },
    {cls:"^Wfica$"                      },
]

# Add remote desktop clients & VMs for no remapping
terminals.extend(remotes)
termStr_ext = toRgxStr(terminals)

terminals_and_remotes = [
    {lst:terminals_lod                  },
    {lst:remotes_lod                    },
]

mscodes.extend(remotes)
mscodeStr_ext = toRgxStr(mscodes)

mscodes_and_remotes = [
    {lst:mscodes_lod                    },
    {lst:remotes_lod                    },
]

# Use for browser specific hotkeys
browsers = [
    "Brave-browser",
    "Chromium",
    "Chromium-browser",
    "Discord",
    "Epiphany",
    "Firefox",
    "Firefox Developer Edition",
    "firefoxdeveloperedition",
    "Google-chrome",
    "microsoft-edge",
    "microsoft-edge-dev",
    "Navigator",
    "org.deepin.browser",
    "Waterfox",
]
browsers = [x.casefold() for x in browsers]
browserStr = "|".join(str('^'+x+'$') for x in browsers)

chromes = [
    "Brave-browser",
    "Chromium",
    "Chromium-browser",
    "Google-chrome",
    "microsoft-edge",
    "microsoft-edge-dev",
    "org.deepin.browser",
]
chromes = [x.casefold() for x in chromes]
chromeStr = "|".join(str('^'+x+'$') for x in chromes)

filemanagers = [
    "caja",
    "dde-file-manager",
    "dolphin",
    "io.elementary.files",
    "nautilus",
    "nemo",
    "org.gnome.nautilus",
    "pcmanfm",
    "pcmanfm-qt",
    "spacefm",
    "thunar",
]
filemanagers = [x.casefold() for x in filemanagers]
filemanagerStr = "|".join(str('^'+x+'$') for x in filemanagers)

### dialogs_Esc = send these windows the Escape key for Cmd+W
dialogs_Esc = [
    {cls:"^.*nautilus$",                nme:"^.*Properties$|^Preferences$|^Create Archive$"},
    {cls:"^Transmission-gtk$",          not_nme:"^Transmission$"},
    {cls:"^org.gnome.Software$",        not_nme:"^Software$"},
]

### dialogs_AltF4 = send these windows Alt+F4 combo for Cmd+W
dialogs_AltF4 = [
    {cls:"^Gnome-control-center$",      not_nme:"^Settings$"},
    {cls:"^gnome-terminal$",            nme:"^Preferences.*$"},
    {cls:"^pcloud$"                     },
    {cls:"^Totem$",                     not_nme:"^Videos$"},
]

# Lists of keyboard device names, to match keyboard type
keyboards_IBM = [
    # Add specific IBM-style keyboard device names to this list
    'IBM Enhanced (101/102-key) Keyboard',
    'IBM Rapid Access Keyboard',
    'IBM Space Saver II',
    'IBM Model M',
    'IBM Model F',    
]
keyboards_Chromebook = [
    # Add specific Chromebook keyboard device names to this list
    'Google .* Keyboard',
    '.* Chromebook Keyboard',
]
keyboards_Windows = [
    # Add specific Windows/PC keyboard device names to this list
    'AT Translated Set 2 keyboard',
]
keyboards_Apple = [
    # Add specific Apple/Mac keyboard device names to this list
    'Mitsumi Electric Apple Extended USB Keyboard',
]

kbtype_lists = {
    'IBM': keyboards_IBM, 
    'Chromebook': keyboards_Chromebook, 
    'Windows': keyboards_Windows, 
    'Apple': keyboards_Apple
}



#################################  MODMAPS  ####################################
###                                                                          ###
###                                                                          ###
###      ███    ███  ██████  ██████  ███    ███  █████  ██████  ███████      ###
###      ████  ████ ██    ██ ██   ██ ████  ████ ██   ██ ██   ██ ██           ###
###      ██ ████ ██ ██    ██ ██   ██ ██ ████ ██ ███████ ██████  ███████      ###
###      ██  ██  ██ ██    ██ ██   ██ ██  ██  ██ ██   ██ ██           ██      ###
###      ██      ██  ██████  ██████  ██      ██ ██   ██ ██      ███████      ###
###                                                                          ###
###                                                                          ###
################################################################################
### Modmaps turn a key into a different key as long as the modmap is active
### The modified key can be used in shortcut combos as the new key


# Change the media_arrows_fix variable to True to enable by default 
# (see above in "VARS"/"VARIABLES" section)
modmap("Cond modmap - Media Arrows Fix",{
    # Fix arrow keys with media functions instead of PgUp/PgDn/Home/End
    Key.PLAYPAUSE:              Key.PAGE_UP,
    Key.STOPCD:                 Key.PAGE_DOWN,
    Key.PREVIOUSSONG:           Key.HOME,
    Key.NEXTSONG:               Key.END,
}, when = lambda _: media_arrows_fix is True)


# Change the mac_numpad variable to False to disable by default 
# (see above in "VARS"/"VARIABLES" section)
modmap("Cond modmap - Mac Numpad feature",{
    # Make numpad be a numpad regardless of Numlock state (like an Apple keyboard in macOS)
    Key.KP1:                    Key.KEY_1,
    Key.KP2:                    Key.KEY_2,
    Key.KP3:                    Key.KEY_3,
    Key.KP4:                    Key.KEY_4,
    Key.KP5:                    Key.KEY_5,
    Key.KP6:                    Key.KEY_6,
    Key.KP7:                    Key.KEY_7,
    Key.KP8:                    Key.KEY_8,
    Key.KP9:                    Key.KEY_9,
    Key.KP0:                    Key.KEY_0,
    Key.KPDOT:                  Key.DOT,  
    Key.KPENTER:                Key.ENTER,
}, when = lambda _: mac_numpad is True)


modmap("Cond modmap - GTK3 numpad nav keys fix",{
    # Make numpad nav keys work correctly in GTK3 apps
    # Key.KP5:                    Key.X,                          # GTK3 numpad fix - TEST TO SEE IF WORKING
    # Numpad PgUp/PgDn/Home/End keys
    Key.KP9:                    Key.PAGE_UP, 
    Key.KP3:                    Key.PAGE_DOWN, 
    Key.KP7:                    Key.HOME, 
    Key.KP1:                    Key.END,
    # Numpad arrow keys
    Key.KP8:                    Key.UP, 
    Key.KP2:                    Key.DOWN, 
    Key.KP4:                    Key.LEFT, 
    Key.KP6:                    Key.RIGHT,
    # Numpad Insert/Delete/Enter keys
    Key.KP0:                    Key.INSERT, 
    Key.KPDOT:                  Key.DELETE, 
    Key.KPENTER:                Key.ENTER,
}, when = lambda ctx: ctx.numlock_on is False and mac_numpad is False)


multipurpose_modmap("Optional Tweaks",
    # {Key.ENTER:                 [Key.ENTER, Key.RIGHT_CTRL]     # Enter2Cmd
    # {Key.CAPSLOCK:              [Key.ESC, Key.RIGHT_CTRL]       # Caps2Esc
    # {Key.LEFT_META:             [Key.ESC, Key.RIGHT_CTRL]       # Caps2Esc - Chromebook
    {                                                             # Placeholder
})


# Fix for avoiding modmapping when using Synergy keyboard/mouse sharing.
# Synergy doesn't set a wm_class, so this may cause issues with other
# applications that also don't set the wm_class.
# Enable only if you use Synergy.
# define_conditional_modmap(lambda wm_class: wm_class == '', {})
###########################
# For use with keyszer, to avoid catching GNOME Shell task switcher: 
# modmap("Synergy fix", {}, when = lambda ctx: ctx.wm_class == '' and ctx.wm_name != '')


# [Global GUI conditional modmaps] Change modifier keys as in xmodmap
modmap("Cond modmap - GUI - Caps2Cmd - not Chromebook kdb", {
    Key.CAPSLOCK:               Key.RIGHT_CTRL,                 # Caps2Cmd
}, when = lambda ctx: matchProps(not_lst=terminals_and_remotes)(ctx) and Caps2Cmd and isKBtype('Chromebook')(ctx) )
modmap("Cond modmap - GUI - Caps2Cmd - Chromebook kdb", {
    Key.LEFT_META:              Key.RIGHT_CTRL,                 # Caps2Cmd - Chromebook
}, when = lambda ctx: matchProps(not_lst=terminals_and_remotes)(ctx) and Caps2Cmd and isKBtype('Chromebook')(ctx) )
modmap("Cond modmap - GUI - IBM kbd", {
    # - IBM
    Key.CAPSLOCK:               Key.LEFT_META,                  # IBM
    Key.LEFT_CTRL:              Key.LEFT_ALT,                   # IBM
    Key.LEFT_ALT:               Key.RIGHT_CTRL,                 # IBM
    Key.RIGHT_ALT:              Key.RIGHT_CTRL,                 # IBM - Multi-language (Remove)
    Key.RIGHT_CTRL:             Key.RIGHT_ALT,                  # IBM - Multi-language (Remove)
}, when = lambda ctx: matchProps(not_lst=terminals_and_remotes)(ctx) and isKBtype('IBM')(ctx) )
modmap("Cond modmap - GUI - Chromebook kbd", {
    # - Chromebook
    Key.LEFT_CTRL:              Key.LEFT_ALT,                   # Chromebook
    Key.LEFT_ALT:               Key.RIGHT_CTRL,                 # Chromebook
    Key.RIGHT_ALT:              Key.RIGHT_CTRL,                 # Chromebook - Multi-language (Remove)
    Key.RIGHT_CTRL:             Key.RIGHT_ALT,                  # Chromebook - Multi-language (Remove)
}, when = lambda ctx: matchProps(not_lst=terminals_and_remotes)(ctx) and isKBtype('Chromebook')(ctx) )
modmap("Cond modmap - GUI - Windows kbd", {
    # - Default Mac/Win
    # - Default Win
    Key.LEFT_CTRL:              Key.LEFT_META,                  # WinMac
    Key.LEFT_META:              Key.LEFT_ALT,                   # WinMac
    Key.LEFT_ALT:               Key.RIGHT_CTRL,                 # WinMac
    Key.RIGHT_ALT:              Key.RIGHT_CTRL,                 # WinMac - Multi-language (Remove)
    Key.RIGHT_META:             Key.RIGHT_ALT,                  # WinMac - Multi-language (Remove)
    Key.RIGHT_CTRL:             Key.RIGHT_META,                 # WinMac - Multi-language (Remove)
}, when = lambda ctx: matchProps(not_lst=terminals_and_remotes)(ctx) and isKBtype('Windows')(ctx) )
modmap("Cond modmap - GUI - Mac kbd", {
    # - Mac Only
    Key.LEFT_CTRL:              Key.LEFT_META,                  # Mac
    Key.LEFT_META:              Key.RIGHT_CTRL,                 # Mac
    Key.RIGHT_META:             Key.RIGHT_CTRL,                 # Mac - Multi-language (Remove)
    Key.RIGHT_CTRL:             Key.RIGHT_META,                 # Mac - Multi-language (Remove)
}, when = lambda ctx: matchProps(not_lst=terminals_and_remotes)(ctx) and isKBtype('Apple')(ctx) )


# [Global Terminals conditional modmaps] Change modifier keys in certain applications
modmap("Cond modmap - Terminals - IBM kbd", {
    # - IBM
    Key.CAPSLOCK:               Key.LEFT_ALT,                   # IBM
    # Left Ctrl Stays Left Ctrl
    Key.LEFT_CTRL:              Key.LEFT_CTRL,                  # IBM
    Key.LEFT_ALT:               Key.RIGHT_CTRL,                 # IBM
    Key.RIGHT_ALT:              Key.RIGHT_CTRL,                 # IBM - Multi-language (Remove)
    # Right Meta does not exist on IBM keyboards
    Key.RIGHT_CTRL:             Key.RIGHT_ALT,                  # IBM

    Key.RIGHT_ALT:              Key.RIGHT_CTRL,                 # IBM - Multi-language (Remove)
    Key.RIGHT_CTRL:             Key.RIGHT_ALT,                  # IBM - Multi-language (Remove)
}, when = lambda ctx: matchProps(lst=terminals_lod)(ctx) and isKBtype('IBM')(ctx) )
modmap("Cond modmap - Terminals - Chromebook kbd", {
    # - Chromebook
    # Left Ctrl Stays Left Ctrl
    Key.LEFT_META:              Key.LEFT_ALT,                   # Chromebook
    Key.LEFT_ALT:               Key.RIGHT_CTRL,                 # Chromebook
    Key.RIGHT_ALT:              Key.RIGHT_CTRL,                 # Chromebook - Multi-language (Remove)
    # Right Meta does not exist on chromebooks
    Key.RIGHT_CTRL:             Key.RIGHT_ALT,                  # Chromebook
}, when = lambda ctx: matchProps(lst=terminals_lod)(ctx) and isKBtype('Chromebook')(ctx) )
modmap("Cond modmap - Terminals - Windows kbd", {
    # - Default Mac/Win
    # - Default Win
    Key.LEFT_CTRL:              Key.LEFT_CTRL,                  # WinMac
    Key.LEFT_META:              Key.LEFT_ALT,                   # WinMac
    Key.LEFT_ALT:               Key.RIGHT_CTRL,                 # WinMac
    Key.RIGHT_ALT:              Key.RIGHT_CTRL,                 # WinMac - Multi-language (Remove)
    Key.RIGHT_META:             Key.RIGHT_ALT,                  # WinMac - Multi-language (Remove)
    Key.RIGHT_CTRL:             Key.LEFT_CTRL,                  # WinMac - Multi-language (Remove)
}, when = lambda ctx: matchProps(lst=terminals_lod)(ctx) and isKBtype('Windows')(ctx) )
modmap("Cond modmap - Terminals - Mac kbd", {
    # - Mac Only
    # Left Ctrl Stays Left Ctrl
    Key.LEFT_META:              Key.RIGHT_CTRL,                 # Mac
    Key.RIGHT_META:             Key.RIGHT_CTRL,                 # Mac - Multi-language (Remove)
    Key.RIGHT_CTRL:             Key.LEFT_CTRL,                  # Mac - Multi-language (Remove)
}, when = lambda ctx: matchProps(lst=terminals_lod)(ctx) and isKBtype('Apple')(ctx) )



###########################################  MAC NUMPAD  #############################################
###                                                                                                ###
###                                                                                                ###
###      ███    ███  █████   ██████     ███    ██ ██    ██ ███    ███ ██████   █████  ██████       ###
###      ████  ████ ██   ██ ██          ████   ██ ██    ██ ████  ████ ██   ██ ██   ██ ██   ██      ###
###      ██ ████ ██ ███████ ██          ██ ██  ██ ██    ██ ██ ████ ██ ██████  ███████ ██   ██      ###
###      ██  ██  ██ ██   ██ ██          ██  ██ ██ ██    ██ ██  ██  ██ ██      ██   ██ ██   ██      ###
###      ██      ██ ██   ██  ██████     ██   ████  ██████  ██      ██ ██      ██   ██ ██████       ###
###                                                                                                ###
###                                                                                                ###
######################################################################################################

# Force the numpad to always be a numpad, like a Mac keyboard on macOS
# Numlock key becomes "Clear" key for use with calculator (sends Escape)
# Toggle feature on/off with Option+Numlock (Fn+Numlock might work on Apple keyboards that have Fn key)
# Set mac_numpad var (above in VARIABLES section) to "False" (no quotes) to disable by default


def mac_numpad_alert():
    """Show notification of state of Toshy's Mac Numpad feature"""
    # Needs "from subprocess import run" somewhere
    if mac_numpad:
        run('notify-send ALERT \
            "Toshy Mac Numpad feature is now ENABLED.\
            \rNumlock becomes "Clear" key (Escape)\
            \rDisable with Option+Numlock."', shell=True)
        debug("Toshy Mac Numpad feature is now ENABLED.")
    # Don't show pointless alert on startup if feature is set to be disabled by default
    if not mac_numpad:
        run('notify-send ALERT \
            "Toshy Mac Numpad feature is now DISABLED.\
            \rRe-enable with Option+Numlock."', shell=True)
        debug("Toshy Mac Numpad feature is now DISABLED.")


def toggle_mac_numpad():
    """Toggle the value of the mac_numpad variable"""
    def _toggle_mac_numpad():
        global mac_numpad
        mac_numpad = not mac_numpad
        mac_numpad_alert()

    return _toggle_mac_numpad


def isNumlockClearKey():
    """Turn the Numlock key into Clear (Esc) key while Mac Numpad feature is enabled."""
    def _isNumlockClearKey():
        global mac_numpad
        if mac_numpad:
            return C("Esc")
        else:
            return C("Numlock")

    return _isNumlockClearKey



###########################  OPTSPECIALCHARS  ##############################
###                                                                      ###
###                                                                      ###
###       ██████  ██████  ████████ ███████ ██████  ███████  ██████       ###
###      ██    ██ ██   ██    ██    ██      ██   ██ ██      ██            ###
###      ██    ██ ██████     ██    ███████ ██████  █████   ██            ###
###      ██    ██ ██         ██         ██ ██      ██      ██            ###
###       ██████  ██         ██    ███████ ██      ███████  ██████       ###
###                                                                      ###
###                                                                      ###
############################################################################

# import the Option key special character entry scheme module (OptSpecialChars)
# include("optspecialchars.py")                   # include file in same folder as config
include("config_includes/optspecialchars.py")   # include file in subfolder



##################################  USER APPS  #######################################
###                                                                                ###
###                                                                                ###
###      ██    ██ ███████ ███████ ██████       █████  ██████  ██████  ███████      ###
###      ██    ██ ██      ██      ██   ██     ██   ██ ██   ██ ██   ██ ██           ###
###      ██    ██ ███████ █████   ██████      ███████ ██████  ██████  ███████      ###
###      ██    ██      ██ ██      ██   ██     ██   ██ ██      ██           ██      ###
###       ██████  ███████ ███████ ██   ██     ██   ██ ██      ██      ███████      ###
###                                                                                ###
###                                                                                ###
######################################################################################

# import custom user config additions from external module file (userapps.py)
# include("userapps.py")                          # include file in same folder as config
include("config_includes/userapps.py")          # include file in subfolder



#################################  MISC APPS  #####################################
###                                                                             ###
###                                                                             ###
###      ███    ███ ██ ███████  ██████      █████  ██████  ██████  ███████      ###
###      ████  ████ ██ ██      ██          ██   ██ ██   ██ ██   ██ ██           ###
###      ██ ████ ██ ██ ███████ ██          ███████ ██████  ██████  ███████      ###
###      ██  ██  ██ ██      ██ ██          ██   ██ ██      ██           ██      ###
###      ██      ██ ██ ███████  ██████     ██   ██ ██      ██      ███████      ###
###                                                                             ###
###                                                                             ###
###################################################################################

keymap("Thunderbird email client", {
    C("Alt-RC-I"):              C("Shift-RC-I"),                # Dev tools
    # Enable Cmd+Option+Left/Right for tab navigation
    C("RC-Alt-Left"):           C("C-Page_Up"),                 # Go to prior tab (macOS Thunderbird tab nav shortcut)
    C("RC-Alt-Right"):          C("C-Page_Down"),               # Go to next tab (macOS Thunderbird tab nav shortcut)
}, when = matchProps(cls="^thunderbird.*$"))

keymap("Angry IP Scanner", {
    C("RC-comma"):              C("Shift-RC-P"),                # Open preferences
}, when = matchProps(cls="^Angry.*IP.*Scanner$"))

keymap("Transmission bittorrent client", {
    C("RC-i"):                  C("Alt-Enter"),                 # Open properties (Get Info) dialog
    C("RC-comma"):             [C("Alt-e"),C("p")],             # Open preferences (settings) dialog
}, when = matchProps(cls="^transmission-gtk$"))

keymap("jDownloader", {
    C("RC-i"):                  C("Alt-Enter"),                 # Open properties
    C("RC-Backspace"):          C("Delete"),                    # Remove download from list
    C("RC-Comma"):              C("C-P"),                       # Open preferences (settings)
}, when = matchProps(cls="^.*jdownloader.*$"))

keymap("Totem video player", {
    C("RC-dot"):                C("C-q"),                       # Stop (quit player, there is no "Stop" function)
}, when = matchProps(cls="^totem$"))

keymap("GNOME image viewer", {
    C("RC-i"):                  C("Alt-Enter"),                 # Image properties
}, when = matchProps(cls="^eog$"))



########################################  FINDER MODS  #############################################
###                                                                                              ###
###                                                                                              ###
###     ███████ ██ ███    ██ ██████  ███████ ██████      ███    ███  ██████  ██████  ███████     ###
###     ██      ██ ████   ██ ██   ██ ██      ██   ██     ████  ████ ██    ██ ██   ██ ██          ###
###     █████   ██ ██ ██  ██ ██   ██ █████   ██████      ██ ████ ██ ██    ██ ██   ██ ███████     ###
###     ██      ██ ██  ██ ██ ██   ██ ██      ██   ██     ██  ██  ██ ██    ██ ██   ██      ██     ###
###     ██      ██ ██   ████ ██████  ███████ ██   ██     ██      ██  ██████  ██████  ███████     ###
###                                                                                              ###
###                                                                                              ###
####################################################################################################

###  START OF FILE MANAGER GROUP OF KEYMAPS - FINDER MODS  ###

# Boolean variable to toggle Enter key state between F2 and Enter
# True = Enter key sends F2, False = Enter key sends Enter
_enter_is_F2 = True # DON'T CHANGE THIS! Must be set to True here. 


def is_Enter_F2(combo_if_true, combo_if_false):
    """
    Send a different combo for Enter key depending on state of _enter_is_F2 variable,\n 
    or latch the variable to True or False to control the Enter key state on next use.
    """
    def _is_Enter_F2():
        global _enter_is_F2
        combo_list = [combo_if_true]
        if combo_if_false in (True, False):                     # Latch variable to given bool value
            _enter_is_F2 = combo_if_false
        elif _enter_is_F2:                                      # If Enter is F2 now, set to be Enter next
            _enter_is_F2 = False
        else:                                                   # If Enter is Enter now, set to be F2 next
            combo_list = [combo_if_false]
            _enter_is_F2 = True
        return combo_list
    return _is_Enter_F2


# Keybindings overrides for Caja
# (overrides some bindings from general file manager code block below)
keymap("Overrides for Caja - Finder Mods", {
    # C("RC-Super-o"):            C("Shift-RC-Enter"),            # Open in new tab
    C("RC-Super-o"):            C("Shift-RC-W"),                # Open in new window
}, when = matchProps(cls="^caja$"))

# Keybindings overrides for DDE (Deepin) File Manager
# (overrides some bindings from general file manager code block below)
keymap("Overrides for DDE File Manager - Finder Mods", {
    C("RC-i"):                  C("RC-i"),                      # File properties dialog (Get Info)
    C("RC-comma"):              None,                           # Disable preferences shortcut (no shortcut available)
    C("RC-Up"):                 C("RC-Up"),                     # Go Up dir
    C("Shift-RC-Left_Brace"):   C("C-Shift-Tab"),               # Go to prior tab
    C("Shift-RC-Right_Brace"):  C("C-Tab"),                     # Go to next tab
    C("Shift-RC-Left"):         C("C-Shift-Tab"),               # Go to prior tab
    C("Shift-RC-Right"):        C("C-Tab"),                     # Go to next tab
}, when = matchProps(cls="^dde-file-manager$"))

# Keybindings overrides for Dolphin (KDE file manager)
# (overrides some bindings from general file manager code block below)
keymap("Overrides for Dolphin - Finder Mods", {
    C("RC-KEY_2"):              C("C-KEY_3"),                   # View as List (Detailed)
    C("RC-KEY_3"):              C("C-KEY_2"),                   # View as List (Compact)
    ##########################################################################################
    ### "Open in new window" (or new tab) requires manually setting custom shortcut of Ctrl+Shift+o
    ### in Dolphin's keyboard shortcuts. There is no default shortcut set for this function.
    ##########################################################################################
    C("RC-Super-o"):            C("Shift-RC-o"),                # Open in new window (or new tab, user's choice, see above)
    C("Shift-RC-N"):            is_Enter_F2(C("F10"), False),   # Create new folder (F10), toggle Enter to be Enter
    C("RC-comma"):              C("Shift-RC-comma"),            # Open preferences dialog
}, when = matchProps(cls="^dolphin$"))

# Keybindings overrides for elementary OS Files (Pantheon)
# (overrides some bindings from general file manager code block below)
keymap("Overrides for Pantheon - Finder Mods", {
    # C("RC-Super-o"):            C("Shift-Enter"),               # Open folder in new tab
    C("RC-comma"):              None,                           # Disable preferences shortcut since none available
}, when = matchProps(cls="^io.elementary.files$"))

# Keybindings overrides for Nautilus
# (overrides some bindings from general file manager code block below)
keymap("Overrides for Nautilus Create Archive dialog - Finder Mods", {
    C("Enter"):                 C("Enter"),                     # Use Enter as Enter in the Create Archive dialog
}, when = matchProps(cls="^.*nautilus$", nme="Create Archive"))

keymap("Overrides for Nautilus - Finder Mods", {
    C("RC-N"):                  C("C-Alt-Space"),               # macOS Finder search window shortcut Cmd+Option+Space
    # For the above shortcut to work, a custom shortcut must be set up in the Settings app in GNOME 
    # to run command: "nautilus --new-window /home/USER" [ replace "USER" ]
    C("RC-KEY_1"):              C("C-KEY_2"),                   # View as Icons
    C("RC-KEY_2"):              C("C-KEY_1"),                   # View as List (Detailed)
    C("RC-Super-o"):            C("Shift-Enter"),               # Open in new window (disable line below)
    # C("RC-Super-o"):            C("RC-Enter"),                  # Open in new tab (disable line above)
    C("RC-comma"):              C("RC-comma"),                  # Overrides "Open preferences dialog" shortcut below
    C("RC-F"):                  C("RC-F"),                      # Don't toggle Enter key, pass Cmd+F
}, when = matchProps(cls="^org.gnome.nautilus$|^nautilus$"))

# Keybindings overrides for PCManFM and PCManFM-Qt
# (overrides some bindings from general file manager code block below)
keymap("Overrides for PCManFM-Qt - Finder Mods", {
    C("RC-Backspace"):          C("Delete"),                    # Move to Trash (delete, bypass dialog)
}, when = matchProps(cls="^pcmanfm-qt$"))

keymap("Overrides for PCManFM - Finder Mods", {
    C("RC-KEY_2"):              C("C-KEY_4"),                   # View as List (Detailed) [Not in PCManFM-Qt]
    C("RC-Backspace"):         [C("Delete"),C("Space")],        # Move to Trash (delete, bypass dialog)
    C("RC-F"):                  C("RC-F"),                      # Don't toggle Enter key state, pass Cmd+F
}, when = matchProps(cls="^pcmanfm$|^pcmanfm-qt$"))

# Keybindings overrides for SpaceFM
# (overrides some bindings from general file manager code block below)
keymap("Overrides for SpaceFM Find Files dialog - Finder Mods", {
    C("Enter"):                 C("Enter"),                     # Use Enter as Enter in the Find dialog
    C("Esc"):                   C("Alt-F4"),                    # Close Find Files dialog with Escape
    C("RC-W"):                  C("Alt-F4"),                    # Close Find Files dialog with Cmd+W
}, when = matchProps(cls="^SpaceFM$", nme="Find FiLes"))

keymap("Overrides for SpaceFM - Finder Mods", {
    C("RC-Page_Up"):            C("C-Shift-Tab"),               # Go to prior tab
    C("RC-Page_Down"):          C("C-Tab"),                     # Go to next tab
    C("Shift-RC-Left_Brace"):   C("C-Shift-Tab"),               # Go to prior tab
    C("Shift-RC-Right_Brace"):  C("C-Tab"),                     # Go to next tab
    C("Shift-RC-Left"):         C("C-Shift-Tab"),               # Go to prior tab
    C("Shift-RC-Right"):        C("C-Tab"),                     # Go to next tab
    C("Shift-RC-N"):    is_Enter_F2(C("RC-F"), False),          # Switch Enter to Enter. New folder is Ctrl+F(???)
    # Need to catch WM_NAME of "Find Files" and override Enter key state back to being Enter. See above keymap.
    C("RC-F"):                  None,                           # No direct shortcut available and menu macros don't work in SpaceFM.
    C("RC-Backspace"):         [C("Delete"),C("Space")],        # Move to Trash (delete, bypass dialog)
    C("RC-comma"):             [C("Alt-V"),C("p")],             # Overrides "Open preferences dialog" shortcut below
    # This shortcut ^^^^^^^^^^^^^^^ is not fully working in SpaceFM. Opens "View" menu but not Preferences.
    # SpaceFM is doing some nasty binding that blocks all shortcuts, including Alt+Tab, while any menu is open.
}, when = matchProps(cls="^spacefm$"))

# Keybindings overrides for Thunar
# (overrides some bindings from general file manager code block below)
keymap("Overrides for Thunar - Finder Mods", {
    C("RC-Super-o"):            C("Shift-RC-P"),                # Open in new tab
    C("RC-comma"):             [C("Alt-E"),C("E")],             # Overrides "Open preferences dialog" shortcut below
    C("RC-F"):                  C("RC-F"),                      # Don't toggle Enter key, pass Cmd+F
}, when = matchProps(cls="^thunar$"))

# Keybindings overrides for GNOME XDG "Save As" and "Open File" dialogs
file_open_save_dialogs = [
    {cls:"^xdg-desktop-portal-gnome$|^firefox.*$", nme:"^Open File$"},
    {cls:"^xdg-desktop-portal-gnome$|^firefox.*$", nme:"^Save As$"},
]
keymap("XDG file dialogs", {
    C("RC-Left"):               C("Alt-Left"),                  # Go Back
    C("RC-Right"):              C("Alt-Right"),                 # Go Forward
    C("RC-Up"):                 C("Alt-Up"),                    # Go Up dir
    C("RC-Down"):               C("Enter"),                     # Go Down dir (open folder/file) [universal]
}, when = matchProps(lst=file_open_save_dialogs))

############################################################
##  Keybindings for Linux general file managers group:
##
##  Currently supported Linux file managers (file browsers):
##  
##  Caja File Browser (MATE file manager, fork of Nautilus)
##  DDE File Manager (Deepin Linux file manager)
##  Dolphin (KDE file manager)
##  Nautilus (GNOME file manager, may be named "Files")
##  Nemo (Cinnamon file manager, fork of Nautilus, may be named "Files")
##  Pantheon Files (elementary OS file manager, may be named "Files")
##  PCManFM (LXDE file manager)
##  PCManFM-Qt (LXQt file manager)
##  SpaceFM (Fork of PCManFM file manager)
##  Thunar File Manager (Xfce file manager)
##  
##  GNOME XDG file dialogs ("Open File" and "Save As" windows in apps like Firefox)
## 
##############################################

keymap("General File Managers - Finder Mods", {
    ###########################################################################################################
    ###  Show Properties (Get Info) | Open Settings/Preferences | Show/Hide hidden files                    ###
    ###########################################################################################################
    C("RC-i"):                  C("Alt-Enter"),                 # File properties dialog (Get Info)
    C("RC-comma"):             [C("Alt-E"),C("N")],             # Open preferences dialog
    C("Shift-RC-dot"):          C("RC-H"),                      # Show/hide hidden files ("dot" files)
    ###########################################################################################################
    ###  Navigation                                                                                         ###
    ###########################################################################################################
    C("RC-Left_Brace"):         C("Alt-Left"),                  # Go Back
    C("RC-Right_Brace"):        C("Alt-Right"),                 # Go Forward
    C("RC-Left"):               C("Alt-Left"),                  # Go Back
    C("RC-Right"):              C("Alt-Right"),                 # Go Forward
    C("RC-Up"):                 C("Alt-Up"),                    # Go Up dir
    # C("RC-Down"):               C("Alt-Down"),                  # Go Down dir (only works on folders) [not universal]
    # C("RC-Down"):               C("RC-O"),                      # Go Down dir (open folder/file) [not universal]
    C("RC-Down"):               C("Enter"),                     # Go Down dir (open folder/file) [universal]
    C("Shift-RC-Left_Brace"):   C("C-Page_Up"),                 # Go to prior tab
    C("Shift-RC-Right_Brace"):  C("C-Page_Down"),               # Go to next tab
    C("Shift-RC-Left"):         C("C-Page_Up"),                 # Go to prior tab
    C("Shift-RC-Right"):        C("C-Page_Down"),               # Go to next tab
    ###########################################################################################################
    ###  Open in New Window | Move to Trash                                                                 ###
    ###########################################################################################################
    C("RC-Super-o"):            C("Shift-RC-o"),                # Open in new window (or tab, depends) [not universal]
    C("RC-Backspace"):          C("Delete"),	                # Move to Trash (delete)
    C("RC-Delete"):             None,                           # Block Ctrl+Delete from performing any action (error in macOS)
    ###########################################################################################################
    ###  ENTER-KEY-TO-RENAME CUSTOM FUNCTION SHORTCUTS                                                      ###
    ###########################################################################################################
    C("Enter"):                 is_Enter_F2(C("F2"),C("Enter")),        # Send F2 to rename files, unless var is False
    C("Shift-RC-N"):            is_Enter_F2(C("Shift-RC-N"), False),    # New folder, set Enter to Enter
    C("RC-L"):                  is_Enter_F2(C("RC-L"), False),          # Set Enter to Enter for Location field
    C("RC-F"):                  is_Enter_F2(C("RC-F"), False),          # Set Enter to Enter for Find field
    C("Esc"):                   is_Enter_F2(C("Esc"), True),            # Send Escape, make sure Enter is back to F2
    C("Shift-RC-Enter"):        C("Enter"),                             # alternative "Enter" key for unusual cases
}, when = matchProps(cls=filemanagerStr))



###################################  BROWSERS  #####################################
###                                                                              ###
###                                                                              ###
###      ██████  ██████   ██████  ██     ██ ███████ ███████ ██████  ███████      ###
###      ██   ██ ██   ██ ██    ██ ██     ██ ██      ██      ██   ██ ██           ###
###      ██████  ██████  ██    ██ ██  █  ██ ███████ █████   ██████  ███████      ###
###      ██   ██ ██   ██ ██    ██ ██ ███ ██      ██ ██      ██   ██      ██      ###
###      ██████  ██   ██  ██████   ███ ███  ███████ ███████ ██   ██ ███████      ###
###                                                                              ###
###                                                                              ###
####################################################################################

# Open preferences in Firefox browsers
keymap("Firefox Browsers Overrides", {
    C("C-comma"):              [C("C-t"),sleep(0.05),ST("about:preferences"),C("Enter")],
    C("Shift-RC-N"):            C("Shift-RC-P"),                      # Open private window with Cmd+Shift+N like other browsers
    K("RC-Backspace"):         [K("Shift-Home"), K("Backspace")],     # Delete Entire Line Left of Cursor
    K("RC-Delete"):            [K("Shift-End"), K("Delete")],         # Delete Entire Line Right of Cursor
}, when = matchProps(cls="^Firefox.*$"))

keymap("Chrome Browsers Overrides", {
    C("C-comma"):              [C("Alt-e"), C("s"),C("Enter")], # Open preferences
    C("RC-q"):                  C("Alt-F4"),                    # Quit Chrome(s) browsers with Cmd+Q
    # C("RC-Left"):               C("Alt-Left"),                  # Page nav: Back to prior page in history (conflict with wordwise)
    # C("RC-Right"):              C("Alt-Right"),                 # Page nav: Forward to next page in history (conflict with wordwise)
    C("RC-Left_Brace"):         C("Alt-Left"),                  # Page nav: Back to prior page in history
    C("RC-Right_Brace"):        C("Alt-Right"),                 # Page nav: Forward to next page in history
}, when = matchProps(cls=chromeStr))
# Opera C-F12

# Keybindings for General Web Browsers
keymap("General Web Browsers", {
    C("RC-Q"):                  C("RC-Q"),                      # Close all browsers Instances
    C("Alt-RC-I"):              C("Shift-RC-I"),                # Dev tools
    C("Alt-RC-J"):              C("Shift-RC-J"),                # Dev tools
    C("RC-Key_1"):              C("Alt-Key_1"),                 # Jump to Tab #1-#8
    C("RC-Key_2"):              C("Alt-Key_2"),
    C("RC-Key_3"):              C("Alt-Key_3"),
    C("RC-Key_4"):              C("Alt-Key_4"),
    C("RC-Key_5"):              C("Alt-Key_5"),
    C("RC-Key_6"):              C("Alt-Key_6"),
    C("RC-Key_7"):              C("Alt-Key_7"),
    C("RC-Key_8"):              C("Alt-Key_8"),
    C("RC-Key_9"):              C("Alt-Key_9"),                 # Jump to last tab
    # Enable Cmd+Shift+Braces for tab navigation (redundant with General GUI)
    # C("Shift-RC-Left_Brace"):   C("C-Page_Up"),                 # Go to prior tab
    # C("Shift-RC-Right_Brace"):  C("C-Page_Down"),               # Go to next tab
    # Enable Cmd+Option+Left/Right for tab navigation
    C("RC-Alt-Left"):           C("C-Page_Up"),                 # Go to prior tab
    C("RC-Alt-Right"):          C("C-Page_Down"),               # Go to next tab
    # Enable Ctrl+PgUp/PgDn for tab navigation
    C("Super-Page_Up"):         C("C-Page_Up"),                 # Go to prior tab
    C("Super-Page_Down"):       C("C-Page_Down"),               # Go to next tab
    # Use Cmd+Braces keys for tab navigation instead of page navigation 
    # C("C-Left_Brace"):        C("C-Page_Up"),
    # C("C-Right_Brace"):       C("C-Page_Down"),
}, when = matchProps(cls=browserStr))



############################################  CODE EDITORS  ##############################################
###                                                                                                    ###
###                                                                                                    ###
###       ██████  ██████  ██████  ███████     ███████ ██████  ██ ████████  ██████  ██████  ███████     ###
###      ██      ██    ██ ██   ██ ██          ██      ██   ██ ██    ██    ██    ██ ██   ██ ██          ###
###      ██      ██    ██ ██   ██ █████       █████   ██   ██ ██    ██    ██    ██ ██████  ███████     ###
###      ██      ██    ██ ██   ██ ██          ██      ██   ██ ██    ██    ██    ██ ██   ██      ██     ###
###       ██████  ██████  ██████  ███████     ███████ ██████  ██    ██     ██████  ██   ██ ███████     ###
###                                                                                                    ###
###                                                                                                    ###
##########################################################################################################

# Keybindings for IntelliJ
keymap("Jetbrains", {
    # General
    C("C-Key_0"):               C("Alt-Key_0"),                 # Open corresponding tool window
    C("C-Key_1"):               C("Alt-Key_1"),                 # Open corresponding tool window
    C("C-Key_2"):               C("Alt-Key_2"),                 # Open corresponding tool window
    C("C-Key_3"):               C("Alt-Key_3"),                 # Open corresponding tool window
    C("C-Key_4"):               C("Alt-Key_4"),                 # Open corresponding tool window
    C("C-Key_5"):               C("Alt-Key_5"),                 # Open corresponding tool window
    C("C-Key_6"):               C("Alt-Key_6"),                 # Open corresponding tool window
    C("C-Key_7"):               C("Alt-Key_7"),                 # Open corresponding tool window
    C("C-Key_8"):               C("Alt-Key_8"),                 # Open corresponding tool window
    C("C-Key_9"):               C("Alt-Key_9"),                 # Open corresponding tool window
    C("Super-Grave"):           C("C-Grave"),                   # Quick switch current scheme
    C("C-Comma"):               C("C-Alt-s"),                   # Open Settings dialog
    C("C-Semicolon"):           C("C-Alt-Shift-s"),             # Open Project Structure dialog
    # Debugging
    C("C-Alt-r"):               C("F9"),                        # Resume program
    # Search/Replace
    C("C-g"):                   C("F3"),                        # Find next
    C("C-Shift-F3"):            C("Shift-F3"),                  # Find previous
    C("Super-g"):               C("Alt-j"),                     # Select next occurrence
    C("C-Super-g"):             C("C-Alt-Shift-j"),             # Select all occurrences
    C("Super-Shift-g"):         C("Alt-Shift-j"),               # Unselect occurrence
    # Editing
    # C("Super-Space"):           C("LC-Space"),                  # Basic code completion
    C("Super-Shift-Space"):     C("Shift-LC-Space"),            # Smart code completion
    C("Super-j"):               C("C-q"),                       # Quick documentation lookup
    C("C-n"):                   C("Alt-Insert"),                # Generate code...
    C("Super-o"):               C("C-o"),                       # Override methods
    C("Super-i"):               C("C-i"),                       # Implement methods
    C("Alt-Up"):                C("C-w"),                       # Extend selection
    C("Alt-Down"):              C("C-Shift-w"),                 # Shrink selection
    C("Super-Shift-q"):         C("Alt-q"),                     # Context info
    C("Super-Alt-o"):           C("C-Alt-o"),                   # Optimize imports
    C("Super-Alt-i"):           C("C-Alt-i"),                   # Auto-indent line(s)
    C("C-Backspace"):           C("C-y"),                       # Delete line at caret
    C("Super-Shift-j"):         C("C-Shift-j"),                 # Smart line join
    C("Alt-Delete"):            C("C-Delete"),                  # Delete to word end
    C("Alt-Backspace"):         C("C-Backspace"),               # Delete to word start
    C("C-Shift-Equal"):         C("C-KPPLUS"),                  # Expand code block
    C("C-Minus"):               C("C-KPMINUS"),                 # Collapse code block
    C("C-Shift-Equal"):         C("C-Shift-KPPLUS"),            # Expand all
    C("C-Shift-Minus"):         C("C-Shift-KPMINUS"),           # Collapse all
    C("C-w"):                   C("C-F4"),                      # Close active editor tab
    # Refactoring
    C("C-Delete"):              C("Alt-Delete"),                # Safe Delete
    C("C-T"):                   C("C-Alt-Shift-t"),             # Refactor this
    # Navigation
    C("C-o"):                   C("C-n"),                       # Go to class
    C("C-Shift-o"):             C("C-Shift-n"),                 # Go to file
    C("C-Alt-o"):               C("C-Alt-Shift-n"),             # Go to symbol
    C("Super-Right"):           C("Alt-Right"),                 # Go to next editor tab
    C("Super-Left"):            C("Alt-Left"),                  # Go to previous editor tab
    C("C-l"):                   C("C-g"),                       # Go to line
    C("Alt-Space"):             C("C-Shift-i"),                 # Open quick definition lookup
    C("C-Y"):                   C("C-Shift-i"),                 # Open quick definition lookup
    C("Super-Shift-b"):         C("C-Shift-b"),                 # Go to type declaration
    C("Super-Up"):              C("Alt-Up"),                    # Go to previous
    C("Super-Down"):            C("Alt-Down"),                  # Go to next method
    C("Super-h"):               C("C-h"),                       # Type hierarchy
    C("Super-Alt-h"):           C("C-Alt-h"),                   # Call hierarchy
    C("C-Down"):                C("C-Enter"),                   # Edit source/View source
    C("Alt-Home"):              C("Alt-Home"),                  # Show navigation bar
    C("F2"):                    C("F11"),                       # Toggle bookmark
    C("Super-F3"):              C("C-F11"),                     # Toggle bookmark with mnemonic
    C("Super-Key_0"):           C("C-Key_0"),                   # Go to numbered bookmark
    C("Super-Key_1"):           C("C-Key_1"),                   # Go to numbered bookmark
    C("Super-Key_2"):           C("C-Key_2"),                   # Go to numbered bookmark
    C("Super-Key_3"):           C("C-Key_3"),                   # Go to numbered bookmark
    C("Super-Key_4"):           C("C-Key_4"),                   # Go to numbered bookmark
    C("Super-Key_5"):           C("C-Key_5"),                   # Go to numbered bookmark
    C("Super-Key_6"):           C("C-Key_6"),                   # Go to numbered bookmark
    C("Super-Key_7"):           C("C-Key_7"),                   # Go to numbered bookmark
    C("Super-Key_8"):           C("C-Key_8"),                   # Go to numbered bookmark
    C("Super-Key_9"):           C("C-Key_9"),                   # Go to numbered bookmark
    C("C-F3"):                  C("Shift-F11"),                 # Show bookmarks
    # Compile and Run
    C("Super-Alt-r"):           C("Alt-Shift-F10"),             # Select configuration and run
    C("Super-Alt-d"):           C("Alt-Shift-F9"),              # Select configuration and debug
    C("Super-r"):               C("Shift-F10"),                 # Run
    C("Super-d"):               C("Shift-F9"),                  # Debug
    C("Super-Shift-r"):         C("C-Shift-F10"),               # Run context configuration from editor
    C("Super-Shift-d"):         C("C-Shift-F9"),                # Debug context configuration from editor
    # VCS/Local History
    C("Super-v"):               C("Alt-Grave"),                 # VCS quick popup
    C("Super-c"):               C("LC-c"),                      # Sigints - interrupt
}, when = matchProps(cls="^jetbrains-(?!.*toolbox).*$"))

keymap("Wordwise - not vscode", {
    # Wordwise remaining - for Everything but VS Code
    C("Alt-Left"):              C("C-Left"),                    # Left of Word
    C("Alt-Shift-Left"):        C("C-Shift-Left"),              # Select Left of Word
    C("Alt-Right"):             C("C-Right"),                   # Right of Word
    C("Alt-Shift-Right"):       C("C-Shift-Right"),             # Select Right of Word
    C("Alt-Shift-g"):           C("C-Shift-g"),                 # View source control
    # ** VS Code fix **
    #   Electron issue precludes normal keybinding fix.
    #   Alt menu auto-focus/toggle gets in the way.
    #
    #   refer to ./xkeysnail-config/vscode_keybindings.json
    # **
    #
    # ** Firefox fix **
    #   User will need to set "ui.key.menuAccessKeyFocuses"
    #   under about:config to false.
    #
    #   https://superuser.com/questions/770301/pentadactyl-how-to-disable-menu-bar-toggle-by-alt
    # **
    #
# }, when = lambda ctx: ctx.wm_class.casefold() not in mscodes)
}, when = matchProps(not_cls=mscodeStr_ext))


# Keybindings for VS Code and variants
keymap("VSCodes overrides for Chromebook/IBM - Sublime", {
    C("C-Alt-g"):               C("C-f2"),                      # Chromebook/IBM - Sublime - find_all_under
}, when = lambda ctx: matchProps(cls=mscodeStr)(ctx) and ( isKBtype('Chromebook')(ctx) or isKBtype('IBM')(ctx) ) and VSCode_Sublime)
keymap("VSCodes overrides for not Chromebook/IBM - Sublime", {
    C("Super-C-g"):             C("C-f2"),                      # Default - Sublime - find_all_under
}, when = lambda ctx: matchProps(cls=mscodeStr)(ctx) and not ( isKBtype('Chromebook')(ctx) or isKBtype('IBM')(ctx) ) and VSCode_Sublime)
keymap("VSCodes overrides for Chromebook/IBM", {
    C("Alt-c"):                 C("LC-c"),                      #  Chromebook/IBM - Terminal - Sigint
    C("Alt-x"):                 C("LC-x"),                      #  Chromebook/IBM - Terminal - Exit nano
}, when = lambda ctx: matchProps(cls=mscodeStr)(ctx) and ( isKBtype('Chromebook')(ctx) or isKBtype('IBM')(ctx) ) )
keymap("VSCodes overrides for not Chromebook/IBM", {
    C("Super-c"):               C("LC-c"),                      # Default - Terminal - Sigint
    C("Super-x"):               C("LC-x"),                      # Default - Terminal - Exit nano
}, when = lambda ctx: matchProps(cls=mscodeStr)(ctx) and not ( isKBtype('Chromebook')(ctx) or isKBtype('IBM')(ctx) ) )
keymap("VSCodes", {
    C("Super-Space"):           C("LC-Space"),                  # Basic code completion (conflicts with input switching)

    # Find dialog options
    C("Alt-RC-C"):              C("Alt-C"),                     # Find: toggle "Match Case"
    C("Alt-RC-W"):              C("Alt-W"),                     # Find: toggle "Match Whole Word"
    C("Alt-RC-R"):              C("Alt-R"),                     # Find: toggle "Use Regular Expression"
    C("Alt-RC-L"):              C("Alt-L"),                     # Find: toggle "Find in Selection"
    C("Alt-RC-P"):              C("Alt-P"),                     # Replace: toggle "Preserve Case"

    C("Alt-RC-Z"):              C("Alt-Z"),                     # View: toggle "Word Wrap"

    # Wordwise remaining - for VS Code
    # Alt-F19 hack fixes Alt menu activation
    C("Alt-Left"):             [C("Alt-F19"),C("C-Left")],          # Left of Word
    C("Alt-Right"):            [C("Alt-F19"),C("C-Right")],         # Right of Word
    C("Alt-Shift-Left"):       [C("Alt-F19"),C("C-Shift-Left")],    # Select Left of Word
    C("Alt-Shift-Right"):      [C("Alt-F19"),C("C-Shift-Right")],   # Select Right of Word

    C("RC-Backspace"):          C("C-Backspace"),               # Delete Entire Line Left of Cursor

    # C("C-PAGE_DOWN"):           ignore_key,                     # cancel next_view
    # C("C-PAGE_UP"):             ignore_key,                     # cancel prev_view
    C("C-Alt-Left"):            C("C-PAGE_UP"),                 # next_view
    C("C-Alt-Right"):           C("C-PAGE_DOWN"),               # prev_view
    C("Shift-RC-Left_Brace"):   C("C-PAGE_UP"),                 # next_view
    C("Shift-RC-Right_Brace"):  C("C-PAGE_DOWN"),               # prev_view

    # VS Code Shortcuts
    C("C-g"):                   ignore_key,                     # cancel Go to Line...
    C("Super-g"):               C("C-g"),                       # Go to Line...
    C("F3"):                    ignore_key,                     # cancel Find next
    C("C-h"):                   ignore_key,                     # cancel replace
    C("C-Alt-f"):               C("C-h"),                       # replace
    C("C-Shift-h"):             ignore_key,                     # cancel replace_next
    C("C-Alt-e"):               C("C-Shift-h"),                 # replace_next
    C("f3"):                    ignore_key,                     # cancel find_next
    C("C-g"):                   C("f3"),                        # find_next
    C("Shift-f3"):              ignore_key,                     # cancel find_prev
    C("C-Shift-g"):             C("Shift-f3"),                  # find_prev
    # C("Super-c"):               C("LC-c"),                      # Default - Terminal - Sigint
    # C("Super-x"):               C("LC-x"),                      # Default - Terminal - Exit nano
    # C("Alt-c"):                 C("LC-c"),                      #  Chromebook/IBM - Terminal - Sigint
    # C("Alt-x"):                 C("LC-x"),                      #  Chromebook/IBM - Terminal - Exit nano
    # C("Super-C-g"):             C("C-f2"),                      # Default - Sublime - find_all_under
    # C("C-Alt-g"):               C("C-f2"),                      # Chromebook/IBM - Sublime - find_all_under
    # C("Super-Shift-up"):        C("Alt-Shift-up"),              # multi-cursor up - Sublime
    # C("Super-Shift-down"):      C("Alt-Shift-down"),            # multi-cursor down - Sublime
    # C(""):                      ignore_key,                     # cancel
    # C(""):                      C(""),                          #
}, when = matchProps(cls=mscodeStr))

# Keybindings for Sublime Text
keymap("Sublime Text overrides for Chromebook/IBM", {
    C("Alt-c"):                 C("LC-c"),                      #  Chromebook/IBM - Terminal - Sigint
    C("Alt-x"):                 C("LC-x"),                      #  Chromebook/IBM - Terminal - Exit nano
    C("Alt-Refresh"):           ignore_key,                     # Chromebook/IBM - cancel find_all_under
    C("Alt-C-g"):               C("Alt-Refresh"),               # Chromebook/IBM - find_all_under
}, when = lambda ctx: matchProps(cls=sublimeStr)(ctx) and ( isKBtype('Chromebook')(ctx) or isKBtype('IBM')(ctx) ) )
keymap("Sublime Text overrides for not Chromebook/IBM", {
    # C("Super-c"):               C("LC-c"),                      # Default - Terminal - Sigint
    # C("Super-x"):               C("LC-x"),                      # Default - Terminal - Exit nano
    C("Alt-f3"):                ignore_key,                     # Default - cancel find_all_under
    C("Super-C-g"):             C("Alt-f3"),                    # Default - find_all_under
}, when = lambda ctx: matchProps(cls=sublimeStr)(ctx) and not ( isKBtype('Chromebook')(ctx) or isKBtype('IBM')(ctx) ) )
keymap("Sublime Text", {
    # C("Super-c"):               C("LC-c"),                      # Default - Terminal - Sigint
    # C("Super-x"):               C("LC-x"),                      # Default - Terminal - Exit nano
    # C("Alt-c"):                 C("LC-c"),                      #  Chromebook/IBM - Terminal - Sigint
    # C("Alt-x"):                 C("LC-x"),                      #  Chromebook/IBM - Terminal - Exit nano
    C("Super-Space"):           C("C-Space"),                   # Basic code completion
    C("C-Super-up"):            C("Alt-o"),                     # Switch file
    C("Super-RC-f"):            C("f11"),                       # toggle_full_screen
    C("C-Alt-v"):              [C("C-k"), C("C-v")],            # paste_from_history
    C("C-up"):                  ignore_key,                     # cancel scroll_lines up
    C("C-Alt-up"):              C("C-up"),                      # scroll_lines up
    C("C-down"):                ignore_key,                     # cancel scroll_lines down
    C("C-Alt-down"):            C("C-down"),                    # scroll_lines down
    C("Super-Shift-up"):        C("Alt-Shift-up"),              # multi-cursor up
    C("Super-Shift-down"):      C("Alt-Shift-down"),            # multi-cursor down
    C("C-PAGE_DOWN"):           ignore_key,                     # cancel next_view
    C("C-PAGE_UP"):             ignore_key,                     # cancel prev_view
    C("C-Shift-left_brace"):    C("C-PAGE_DOWN"),               # next_view
    C("C-Shift-right_brace"):   C("C-PAGE_UP"),                 # prev_view
    C("C-Alt-right"):           C("C-PAGE_DOWN"),               # next_view
    C("C-Alt-left"):            C("C-PAGE_UP"),                 # prev_view
    C("insert"):                ignore_key,                     # cancel toggle_overwrite
    C("C-Alt-o"):               C("insert"),                    # toggle_overwrite
    C("Alt-c"):                 ignore_key,                     # cancel toggle_case_sensitive
    C("C-Alt-c"):               C("Alt-c"),                     # toggle_case_sensitive
    C("C-h"):                   ignore_key,                     # cancel replace
    C("C-Alt-f"):               C("C-h"),                       # replace
    C("C-Shift-h"):             ignore_key,                     # cancel replace_next
    C("C-Alt-e"):               C("C-Shift-h"),                 # replace_next
    C("f3"):                    ignore_key,                     # cancel find_next
    C("C-g"):                   C("f3"),                        # find_next
    C("Shift-f3"):              ignore_key,                     # cancel find_prev
    C("C-Shift-g"):             C("Shift-f3"),                  # find_prev
    C("C-f3"):                  ignore_key,                     # cancel find_under
    C("Super-Alt-g"):           C("C-f3"),                      # find_under
    C("C-Shift-f3"):            ignore_key,                     # cancel find_under_prev
    C("Super-Alt-Shift-g"):     C("C-Shift-f3"),                # find_under_prev
    C("Alt-f3"):                ignore_key,                     # Default - cancel find_all_under
    C("Super-C-g"):             C("Alt-f3"),                    # Default - find_all_under
    # C("Alt-Refresh"):           ignore_key,                     # Chromebook/IBM - cancel find_all_under
    # C("Alt-C-g"):               C("Alt-Refresh"),               # Chromebook/IBM - find_all_under
    C("C-Shift-up"):            ignore_key,                     # cancel swap_line_up
    C("Super-Alt-up"):          C("C-Shift-up"),                # swap_line_up
    C("C-Shift-down"):          ignore_key,                     # cancel swap_line_down
    C("Super-Alt-down"):        C("C-Shift-down"),              # swap_line_down
    C("C-Pause"):               ignore_key,                     # cancel cancel_build
    C("Super-c"):               C("C-Pause"),                   # cancel_build
    C("f9"):                    ignore_key,                     # cancel sort_lines case_s false
    C("f5"):                    C("f9"),                        # sort_lines case_s false
    C("Super-f9"):              ignore_key,                     # cancel sort_lines case_s true
    C("Super-f5"):              C("Super-f9"),                  # sort_lines case_s true
    C("Alt-Shift-Key_1"):       ignore_key,                     # cancel set_layout
    C("C-Alt-Key_1"):           C("Alt-Shift-Key_1"),           # set_layout
    C("Alt-Shift-Key_2"):       ignore_key,                     # cancel set_layout
    C("C-Alt-Key_2"):           C("Alt-Shift-Key_2"),           # set_layout
    C("Alt-Shift-Key_3"):       ignore_key,                     # cancel set_layout
    C("C-Alt-Key_3"):           C("Alt-Shift-Key_3"),           # set_layout
    C("Alt-Shift-Key_4"):       ignore_key,                     # cancel set_layout
    C("C-Alt-Key_4"):           C("Alt-Shift-Key_4"),           # set_layout
    C("Alt-Shift-Key_8"):       ignore_key,                     # cancel set_layout
    C("C-Alt-Shift-Key_2"):     C("Alt-Shift-Key_8"),           # set_layout
    C("Alt-Shift-Key_9"):       ignore_key,                     # cancel set_layout
    C("C-Alt-Shift-Key_3"):     C("Alt-Shift-Key_9"),           # set_layout
    C("Alt-Shift-Key_5"):       ignore_key,                     # cancel set_layout
    C("C-Alt-Shift-Key_5"):     C("Alt-Shift-Key_5"),           # set_layout
    # C(""):                    ignore_key,                     # cancel
    # C(""):                    C(""),                          #
}, when = matchProps(cls=sublimeStr))



###########################  DIALOG FIXES  ###########################
###                                                                ###
###                                                                ###
###      ██████  ██  █████  ██       ██████   ██████  ███████      ###
###      ██   ██ ██ ██   ██ ██      ██    ██ ██       ██           ###
###      ██   ██ ██ ███████ ██      ██    ██ ██   ███ ███████      ###
###      ██   ██ ██ ██   ██ ██      ██    ██ ██    ██      ██      ###
###      ██████  ██ ██   ██ ███████  ██████   ██████  ███████      ###
###                                                                ###
###                                                                ###
######################################################################
### Fixes for the problem of modal dialogs and other "child" 
### windows failing to close with Cmd+W.
### Many dialogs respond to the Escape key, others may 
### require Alt+F4 to close.
### 
### Cmd+W can't just be always mapped to Alt+F4 for all apps because 
### some apps will "quit" rather than just closing a tab.
### 
### To add window conditions to the list, search for the list names in 
### the "LISTS" section near the top of this config file. 
### dialogs_Esc = send these windows the Escape key for Cmd+W
### dialogs_AltF4 = send these windows Alt+F4 for Cmd+W

keymap("Cmd+W dialog fix - send Escape", {
    C("RC-W"):                  C("Esc"),
}, when = matchProps(lst=dialogs_Esc))

keymap("Cmd+W dialog fix - send Alt+F4", {
    C("RC-W"):                  C("Alt-F4"),
}, when = matchProps(lst=dialogs_AltF4))



###########################  TAB NAV FIXES  #############################
###                                                                   ###
###                                                                   ###
###      ████████  █████  ██████      ███    ██  █████  ██    ██      ###
###         ██    ██   ██ ██   ██     ████   ██ ██   ██ ██    ██      ###
###         ██    ███████ ██████      ██ ██  ██ ███████ ██    ██      ###
###         ██    ██   ██ ██   ██     ██  ██ ██ ██   ██  ██  ██       ###
###         ██    ██   ██ ██████      ██   ████ ██   ██   ████        ###
###                                                                   ###
###                                                                   ###
#########################################################################
### Various fixes for supporting tab navigation shortcuts like Shift+Cmd+Braces

tabUIfix_CtrlShiftTab = [
    {cls:"^deepin-terminal$"},
    {cls:"^.*jDownloader.*$"},
    {cls:"^Kgx$"},
    {cls:"^kitty$"},
    {cls:"^org.gnome.Console$|^Console$"},
]

tabUIfix_CtrlAltPgUp = [
    {cls:"^gedit$"},
]

# Tab navigation overrides for tabbed UI apps that use Ctrl+Shift+Tab/Ctrl+Tab instead of Ctrl+PgUp/PgDn
keymap("Tab Nav fix for apps that use Ctrl+Shift+Tab/Ctrl+Tab", {
    C("Shift-RC-Left_Brace"):   C("C-Shift-Tab"),               # Tab nav: Go to prior tab (left)
    C("Shift-RC-Right_Brace"):  C("C-Tab"),                     # Tab nav: Go to next tab (right)
    C("Shift-RC-Left"):         C("C-Shift-Tab"),               # Tab nav: Go to prior tab (left)
    C("Shift-RC-Right"):        C("C-Tab"),                     # Tab nav: Go to next tab (right)
}, when = matchProps(lst=tabUIfix_CtrlShiftTab))

# Tab navigation overrides for tabbed UI apps that use Ctrl+Alt+PgUp/PgDn instead of Ctrl+PgUp/PgDn
keymap("Tab Nav fix for apps that use Ctrl+Alt+PgUp/PgDn", {
    C("Shift-RC-Left_Brace"):   C("C-Alt-Page_Up"),             # Go to prior tab (Left)
    C("Shift-RC-Right_Brace"):  C("C-Alt-Page_Down"),           # Go to next tab (Right)
}, when = matchProps(lst=tabUIfix_CtrlAltPgUp))

keymap("Konsole tab switching", {
    # Ctrl Tab - In App Tab Switching
    C("LC-Tab") :               C("Shift-Right"),
    C("Shift-LC-Tab") :         C("Shift-Left"),
    C("LC-Grave") :             C("Shift-Left"),
}, when = matchProps(cls="^konsole$"))

keymap("Elementary Terminal tab switching", {
    # Ctrl Tab - In App Tab Switching
    C("LC-Tab") :               C("Shift-LC-Right"),
    C("Shift-LC-Tab") :         C("Shift-LC-Left"),
    C("LC-Grave") :             C("Shift-LC-Left"),
}, when = matchProps(cls="^Io.elementary.terminal$|^kitty$"))



######################################  TERMINALS  #######################################
###                                                                                    ###
###                                                                                    ###
###      ████████ ███████ ██████  ███    ███ ██ ███    ██  █████  ██      ███████      ###
###         ██    ██      ██   ██ ████  ████ ██ ████   ██ ██   ██ ██      ██           ###
###         ██    █████   ██████  ██ ████ ██ ██ ██ ██  ██ ███████ ██      ███████      ###
###         ██    ██      ██   ██ ██  ██  ██ ██ ██  ██ ██ ██   ██ ██           ██      ###
###         ██    ███████ ██   ██ ██      ██ ██ ██   ████ ██   ██ ███████ ███████      ###
###                                                                                    ###
###                                                                                    ###
##########################################################################################

keymap("Alacritty terminal", {
    C("RC-K"):                  C("C-L"),                       # clear log
}, when = matchProps(cls="^alacritty$"))

keymap("Deepin Terminal overrides", {
    C("RC-w"):                  C("Alt-w"),                     # Close only current tab, instead of all other tabs
    C("RC-j"):                  None,                           # Block Cmd+J from remapping to vertical split (Ctrl+Shift+J) 
    C("RC-minus"):              C("C-minus"),                   # Decrease font size/zoom out 
    C("RC-equal"):              C("C-equal"),                   # Increase font size/zoom in
}, when = matchProps(cls="^deepin-terminal$"))

keymap("Kitty terminal - not tab nav", {
    C("RC-L"):                  C("C-L"),                       # Clear log
    C("RC-K"):                  C("C-L"),                       # Clear log (macOS)
}, when = matchProps(cls="^kitty$"))


# Overrides to General Terminals shortcuts for specific distros
keymap("General Terminals overrides - elementary OS", {
    C("LC-Right"):              C("Super-Right"),               # SL - Change workspace eos
    C("LC-Left"):               C("Super-Left"),                # SL - Change workspace eos
}, when = lambda ctx: matchProps(cls=termStr)(ctx) and DISTRO_NAME == 'eos')
keymap("General Terminals overrides - Pop!_OS", {
    C("LC-Right"):              C("Super-C-Up"),                # SL - Change workspace (popos)
    C("LC-Left"):               C("Super-C-Down"),              # SL - Change workspace (popos)
}, when = lambda ctx: matchProps(cls=termStr)(ctx) and DISTRO_NAME == 'popos')
keymap("General Terminals overrides - Ubuntu/Fedora", {
    C("LC-Right"):              C("Super-Page_Up"),             # SL - Change workspace (ubuntu/fedora)
    C("LC-Left"):               C("Super-Page_Down"),           # SL - Change workspace (ubuntu/fedora)
}, when = lambda ctx: matchProps(cls=termStr)(ctx) and ( DISTRO_NAME == 'ubuntu' or DISTRO_NAME == 'fedora' ) )


# Overrides to General Terminals shortcuts for specific desktop environments
keymap("General Terminals overrides - Budgie", {
    C("LC-Right"):              C("C-Alt-Right"),               # Default SL - Change workspace (budgie)
    C("LC-Left"):               C("C-Alt-Left"),                # Default SL - Change workspace (budgie)
}, when = lambda ctx: matchProps(cls=termStr)(ctx) and DESKTOP_ENV == 'budgie' )
keymap("General Terminals overrides - GNOME", {
    ### Keyboard input source (language/layout) switching in GNOME
    C("LC-Space"):              [bind,C("Super-Space")],        # keyboard input source (language) switching (gnome)
    C("Shift-LC-Space"):        [bind,C("Super-Shift-Space")],  # keyboard input source (language) switching (reverse) (gnome)
}, when = lambda ctx: matchProps(cls=termStr)(ctx) and DESKTOP_ENV == 'gnome' )
keymap("General Terminals overrides - Xfce4", {
    C("RC-Grave"):             [bind,C("Super-Tab")],                 # xfce4 Switch within app group
    C("Shift-RC-Grave"):       [bind,C("Super-Shift-Tab")],           # xfce4 Switch within app group
    C("LC-Right"):              C("C-Alt-Home"),                # SL - Change workspace xfce4
    C("LC-Left"):               C("C-Alt-End"),                 # SL - Change workspace xfce4
}, when = lambda ctx: matchProps(cls=termStr)(ctx) and DESKTOP_ENV == 'xfce' )


keymap("General Terminals", {
    ### wordwise overrides of general GUI block
    C("Alt-Backspace"):         C("Alt-Shift-Backspace"),       # Wordwise delete word left of cursor in terminals
    C("Alt-Delete"):           [C("Esc"),C("d")],               # Wordwise delete word right of cursor in terminals
    C("RC-Backspace"):          C("C-u"),                       # Wordwise delete line left of cursor in terminals
    C("RC-Delete"):             C("C-k"),                       # Wordwise delete line right of cursor in terminals
    ### Tab navigation
    C("Shift-RC-Left"):         C("C-Page_Up"),                 # Tab nav: Go to prior tab (Left)
    C("Shift-RC-Right"):        C("C-Page_Down"),               # Tab nav: Go to next tab (Right)
    ### Keyboard input source (language/layout) switching in GNOME 
    C("LC-Space"):              [bind,C("Super-Space")],        # keyboard input source (language) switching (gnome)
    C("Shift-LC-Space"):        [bind,C("Super-Shift-Space")],  # keyboard input source (language) switching (reverse) (gnome)
    C("LC-RC-f"):               C("Alt-F10"),                   # Toggle window maximized state
    # C("RC-Grave"):              C("Super-Tab"),                 # xfce4 Switch within app group
    # C("Shift-RC-Grave"):        C("Super-Shift-Tab"),           # xfce4 Switch within app group
    # C("LC-Right"):              C("C-Alt-Right"),               # Default SL - Change workspace (budgie)
    # C("LC-Left"):               C("C-Alt-Left"),                # Default SL - Change workspace (budgie)
    # C("LC-Right"):              C("C-Alt-Home"),                # SL - Change workspace xfce4
    # C("LC-Left"):               C("C-Alt-End"),                 # SL - Change workspace xfce4
    # C("LC-Right"):              C("Super-Right"),               # SL - Change workspace eos
    # C("LC-Left"):               C("Super-Left"),                # SL - Change workspace eos
    # C("LC-Right"):              C("Super-Page_Up"),             # SL - Change workspace (ubuntu/fedora)
    # C("LC-Left"):               C("Super-Page_Down"),           # SL - Change workspace (ubuntu/fedora)
    # C("LC-Right"):              C("Super-C-Up"),                # SL - Change workspace (popos)
    # C("LC-Left"):               C("Super-C-Down"),              # SL - Change workspace (popos)
    # Ctrl Tab - In App Tab Switching
    C("LC-Tab") :               C("LC-PAGE_DOWN"),
    C("Shift-LC-Tab") :         C("LC-PAGE_UP"),
    C("LC-Grave") :             C("LC-PAGE_UP"),
    # C("Alt-Tab"):               ignore_key,                     # Default - Cmd Tab - App Switching Default
    # C("RC-Tab"):                C("Alt-Tab"),                   # Default - Cmd Tab - App Switching Default
    # C("Shift-RC-Tab"):          C("Alt-Shift-Tab"),             # Default - Cmd Tab - App Switching Default
    # Converts Cmd to use Ctrl-Shift
    C("RC-MINUS"):              C("C-MINUS"),
    C("RC-EQUAL"):              C("C-Shift-EQUAL"),
    # C("RC-BACKSPACE"):          C("C-Shift-BACKSPACE"),           # Conflicts with wordwise shortcut above
    C("RC-W"):                  C("C-Shift-W"),
    C("RC-E"):                  C("C-Shift-E"),
    C("RC-R"):                  C("C-Shift-R"),
    C("RC-T"):                  C("C-Shift-t"),
    C("RC-Y"):                  C("C-Shift-Y"),
    C("RC-U"):                  C("C-Shift-U"),
    C("RC-I"):                  C("C-Shift-I"),
    C("RC-O"):                  C("C-Shift-O"),
    C("RC-P"):                  C("C-Shift-P"),
    C("RC-LEFT_BRACE"):         C("C-Shift-LEFT_BRACE"),
    C("RC-RIGHT_BRACE"):        C("C-Shift-RIGHT_BRACE"),
    C("RC-A"):                  C("C-Shift-A"),
    C("RC-S"):                  C("C-Shift-S"),
    C("RC-D"):                  C("C-Shift-D"),
    C("RC-F"):                  C("C-Shift-F"),
    C("RC-G"):                  C("C-Shift-G"),
    C("RC-H"):                  C("C-Shift-H"),
    C("RC-J"):                  C("C-Shift-J"),
    C("RC-K"):                  C("C-Shift-K"),
    C("RC-L"):                  C("C-Shift-L"),
    C("RC-SEMICOLON"):          C("C-Shift-SEMICOLON"),
    C("RC-APOSTROPHE"):         C("C-Shift-APOSTROPHE"),
    # C("RC-GRAVE"):              C("C-Shift-GRAVE"),             # Conflicts with General GUI window switching
    C("RC-Z"):                  C("C-Shift-Z"),
    C("RC-X"):                  C("C-Shift-X"),
    C("RC-C"):                  C("C-Shift-C"),
    C("RC-V"):                  C("C-Shift-V"),
    C("RC-B"):                  C("C-Shift-B"),
    C("RC-N"):                  C("C-Shift-N"),
    C("RC-M"):                  C("C-Shift-M"),
    C("RC-COMMA"):              C("C-Shift-COMMA"),
    C("RC-Dot"):                C("LC-c"),                      # Mimic macOS Cmd+Dot to cancel command
    C("RC-SLASH"):              C("C-Shift-SLASH"),
    C("RC-KPASTERISK"):         C("C-Shift-KPASTERISK"),

}, when = matchProps(cls=termStr))



#############################  GENERAL GUI  ################################
###                                                                      ###
###                                                                      ###
###       ██████  ███████ ███    ██ ███████ ██████   █████  ██           ###
###      ██       ██      ████   ██ ██      ██   ██ ██   ██ ██           ###
###      ██   ███ █████   ██ ██  ██ █████   ██████  ███████ ██           ###
###      ██    ██ ██      ██  ██ ██ ██      ██   ██ ██   ██ ██           ###
###       ██████  ███████ ██   ████ ███████ ██   ██ ██   ██ ███████      ###
###                                                                      ###
###                                                                      ###
############################################################################

# Note: terminals extends to remotes as well
keymap("Cmd+Dot not in terminals", {
    C("RC-Dot"):                C("Esc"),                       # Mimic macOS Cmd+dot = Escape key (not in terminals)
# }, when = lambda ctx: ctx.wm_class.casefold() not in terminals)
}, when = matchProps(not_cls=termStr_ext) )


# Overrides to General GUI shortcuts for specific keyboard types
keymap("General GUI overrides - Chromebook/IBM", {
    # In-App Tab switching
    C("Alt-Tab"):              [bind,C("C-Tab")],               # Chromebook/IBM - In-App Tab switching
    C("Alt-Shift-Tab"):        [bind,C("C-Shift-Tab")],         # Chromebook/IBM - In-App Tab switching
    C("Alt-Grave") :           [bind,C("C-Shift-Tab")],         # Chromebook/IBM - In-App Tab switching
    C("RAlt-Backspace"):        C("Delete"),                    # Chromebook/IBM - Delete
    C("LAlt-Backspace"):        C("C-Backspace"),               # Chromebook/IBM - Delete Left Word of Cursor
}, when = lambda ctx: matchProps(not_lst=remotes_lod)(ctx) and ( isKBtype('Chromebook')(ctx) or isKBtype('IBM')(ctx) ) )
keymap("General GUI overrides - not Chromebook", {
    # In-App Tab switching
    C("Super-Tab"):            [bind,C("LC-Tab")],              # Default not-chromebook
    C("Super-Shift-Tab"):      [bind,C("Shift-LC-Tab")],        # Default not-chromebook
    C("Alt-Backspace"):         C("C-Backspace"),               # Default not-chromebook
}, when = lambda ctx: matchProps(not_lst=remotes_lod)(ctx) and not isKBtype('Chromebook')(ctx) )


# Overrides to General GUI shortcuts for specific distros
keymap("General GUI overrides - elementary OS", {
    C("RC-F3"):                 C("Super-d"),                   # Default SL - Show Desktop (gnome/kde,eos)
    C("RC-Space"):              C("Super-Space"),               # SL - Launch Application Menu (eos)
    C("RC-LC-f"):               C("Super-Up"),                  # SL- Maximize app eos
}, when = lambda ctx: matchProps(not_lst=remotes_lod)(ctx) and DISTRO_NAME == 'eos' )
keymap("General GUI overrides - Fedora", {
    C("RC-H"):                  C("Super-h"),                   # Default SL - Minimize app (gnome/budgie/popos/fedora) not-deepin
    C("Super-Right"):           C("Super-Page_Up"),             # SL - Change workspace (ubuntu/fedora)
    C("Super-Left"):            C("Super-Page_Down"),           # SL - Change workspace (ubuntu/fedora)
}, when = lambda ctx: matchProps(not_lst=remotes_lod)(ctx) and DISTRO_NAME == 'fedora' )
keymap("General GUI overrides - Manjaro", {
    C("RC-LC-f"):               C("Super-PAGE_UP"),             # SL- Maximize app manjaro
    C("RC-LC-f"):               C("Super-PAGE_DOWN"),           # SL - Minimize app manjaro
}, when = lambda ctx: matchProps(not_lst=remotes_lod)(ctx) and DISTRO_NAME == 'manjaro' )
keymap("General GUI overrides - KDE Neon", {
    C("RC-Super-f"):            C("Super-Page_Up"),             # SL - Toggle maximized window state (kde_neon)
    C("RC-H"):                  C("Super-Page_Down"),           # SL - Minimize app (kde_neon)
                                                                # SL - Default SL - Change workspace (kde_neon)
}, when = lambda ctx: matchProps(not_lst=remotes_lod)(ctx) and DISTRO_NAME == 'neon' )
keymap("General GUI overrides - Pop!_OS", {
    C("RC-H"):                  C("Super-h"),                   # Default SL - Minimize app (gnome/budgie/popos/fedora) not-deepin
    C("Super-Right"):           C("Super-C-Up"),                # SL - Change workspace (popos)
    C("Super-Left"):            C("Super-C-Down"),              # SL - Change workspace (popos)
    C("RC-Q"):                  C("Super-q"),                   # SL - Close Apps (popos)
}, when = lambda ctx: matchProps(not_lst=remotes_lod)(ctx) and DISTRO_NAME == 'popos' )
keymap("General GUI overrides - Ubuntu", {
    C("Super-Right"):           C("Super-Page_Up"),             # SL - Change workspace (ubuntu/fedora)
    C("Super-Left"):            C("Super-Page_Down"),           # SL - Change workspace (ubuntu/fedora)
}, when = lambda ctx: matchProps(not_lst=remotes_lod)(ctx) and DISTRO_NAME == 'ubuntu' )


# Overrides to General GUI shortcuts for specific desktop environments
keymap("General GUI overrides - Budgie", {
    C("Super-Right"):           C("C-Alt-Right"),               # Default SL - Change workspace (budgie)
    C("Super-Left"):            C("C-Alt-Left"),                # Default SL - Change workspace (budgie)
    C("RC-H"):                  C("Super-h"),                   # Default SL - Minimize app (gnome/budgie/popos/fedora) not-deepin
}, when = lambda ctx: matchProps(not_lst=remotes_lod)(ctx) and DESKTOP_ENV == 'budgie' )
keymap("General GUI overrides - Deepin", {
    C("RC-H"):                  C("Super-n"),                   # Default SL - Minimize app (deepin)
    C("Alt-RC-Space"):          C("Super-e"),                   # Open Finder - Placeholder (deepin)
}, when = lambda ctx: matchProps(not_lst=remotes_lod)(ctx) and DESKTOP_ENV == 'deepin' )
keymap("General GUI overrides - GNOME", {
    C("RC-Space"):              C("Alt-F1"),                    # Default SL - Launch Application Menu (gnome/kde)
    C("RC-F3"):                 C("Super-d"),                   # Default SL - Show Desktop (gnome/kde,eos)
    C("RC-Super-f"):            C("Alt-F10"),                   # Default SL - Maximize app (gnome/kde)
    C("RC-H"):                  C("Super-h"),                   # Default SL - Minimize app (gnome/budgie/popos/fedora) not-deepin
}, when = lambda ctx: matchProps(not_lst=remotes_lod)(ctx) and DESKTOP_ENV == 'gnome' )
keymap("General GUI overrides - KDE", {
    C("RC-Space"):              C("Alt-F1"),                    # Default SL - Launch Application Menu (gnome/kde)
    C("RC-F3"):                 C("Super-d"),                   # Default SL - Show Desktop (gnome/kde,eos)
    C("RC-Super-f"):            C("Alt-F10"),                   # Default SL - Maximize app (gnome/kde)
}, when = lambda ctx: matchProps(not_lst=remotes_lod)(ctx) and DESKTOP_ENV == 'kde' )
keymap("General GUI overrides - Xfce4", {
    C("RC-Grave"):             [bind,C("Super-Tab")],           # xfce4 Switch within app group
    C("Shift-RC-Grave"):       [bind,C("Super-Shift-Tab")],     # xfce4 Switch within app group
    C("RC-Space"):              C("LC-Esc"),                    # SL- Launch Application Menu xfce4
    C("RC-F3"):                 C("C-Alt-d"),                   # SL- Show Desktop xfce4
    C("RC-H"):                  C("Alt-F9"),                    # SL - Minimize app xfce4
}, when = lambda ctx: matchProps(not_lst=remotes_lod)(ctx) and DESKTOP_ENV == 'xfce4' )


# None referenced here originally
# - but remote clients and VM software ought to be set here
# These are the typical remaps for ALL GUI based apps
keymap("General GUI", {

    C("Shift-Alt-RC-M"):        toggle_media_arrows_fix(),      # Enable/disable the fix for media functions on arrow keys

    C("Alt-Numlock"):           toggle_mac_numpad(),            # Turn the Mac Numpad feature on and off
    C("Fn-Numlock"):            toggle_mac_numpad(),            # Turn the Mac Numpad feature on and off
    C("Numlock"):               isNumlockClearKey(),            # Turn Numlock key into "Clear" (Esc) if mac_numpad enabled

    C("Shift-RC-Left_Brace"):   C("C-Page_Up"),                 # Tab nav: Go to prior (left) tab
    C("Shift-RC-Right_Brace"):  C("C-Page_Down"),               # Tab nav: Go to next (right) tab
    C("RC-Space"):              C("Alt-F1"),                    # Default SL - Launch Application Menu (gnome/kde)
    C("RC-F3"):                 C("Super-d"),                   # Default SL - Show Desktop (gnome/kde,eos)
    C("RC-Super-f"):            C("Alt-F10"),                   # Default SL - Maximize app (gnome/kde)
    # C("RC-Super-f"):            C("Super-Page_Up"),             # SL - Toggle maximized window state (kde_neon)
    # C("Super-Right"):           C("C-Alt-Right"),               # Default SL - Change workspace (budgie)
    # C("Super-Left"):            C("C-Alt-Left"),                # Default SL - Change workspace (budgie)
    C("RC-Q"):                  C("Alt-F4"),                    # Default SL - not-popos
    C("RC-H"):                  C("Super-h"),                   # Default SL - Minimize app (gnome/budgie/popos/fedora) not-deepin
    # C("RC-H"):                  C("Super-n"),                   # Default SL - Minimize app (deepin)
    # C("Alt-Tab"):               ignore_key,                     # Default - Cmd Tab - App Switching Default
    C("RC-Tab"):               [bind, C("Alt-Tab")],            # Default - Cmd Tab - App Switching Default
    C("Shift-RC-Tab"):         [bind, C("Alt-Shift-Tab")],      # Default - Cmd Tab - App Switching Default
    C("RC-Grave"):             [bind, C("Alt-Grave")],          # Default not-xfce4 - Cmd ` - Same App Switching
    C("Shift-RC-Grave"):       [bind, C("Alt-Shift-Grave")],    # Default not-xfce4 - Cmd ` - Same App Switching
    # C("RC-Grave"):              C("Super-Tab"),                 # xfce4 Switch within app group
    # C("Shift-RC-Grave"):        C("Super-Shift-Tab"),           # xfce4 Switch within app group
    # C("Super-Right"):           C("Super-Page_Up"),             # SL - Change workspace (ubuntu/fedora)
    # C("Super-Left"):            C("Super-Page_Down"),           # SL - Change workspace (ubuntu/fedora)
    # C("Super-Right"):           C("Super-C-Up"),                # SL - Change workspace (popos)
    # C("Super-Left"):            C("Super-C-Down"),              # SL - Change workspace (popos)
    # C("RC-Q"):                  C("Super-q"),                   # SL - Close Apps (popos)
    # C("RC-Space"):              C("Super-Space"),               # SL - Launch Application Menu (eos)
    # C("RC-H"):                  C("Super-Page_Down"),           # SL - Minimize app (kde_neon)
    #                                                             # SL - Default SL - Change workspace (kde_neon)
    # C("RC-Space"):              C("LC-Esc"),                    # SL- Launch Application Menu xfce4
    # C("RC-F3"):                 C("C-Alt-d"),                   # SL- Show Desktop xfce4
    # C("RC-LC-f"):               C("Super-Up"),                  # SL- Maximize app eos
    # C("RC-LC-f"):               C("Super-PAGE_UP"),             # SL- Maximize app manjaro

    # Basic App hotkey functions
    # C("RC-H"):                  C("Alt-F9"),                    # SL - Minimize app xfce4
    # C("RC-LC-f"):               C("Super-PAGE_DOWN"),           # SL - Minimize app manjaro

    # In-App Tab switching
    # C("Alt-Tab"):               C("C-Tab"),                     # Chromebook/IBM - In-App Tab switching
    # C("Alt-Shift-Tab"):         C("C-Shift-Tab"),               # Chromebook/IBM - In-App Tab switching
    # C("Alt-Grave") :            C("C-Shift-Tab"),               # Chromebook/IBM - In-App Tab switching
    # C("Super-Tab"):             C("LC-Tab"),                    # Default not-chromebook
    # C("Super-Shift-Tab"):       C("Shift-LC-Tab"),              # Default not-chromebook

    # Fn to Alt style remaps
    C("RAlt-Enter"):            C("insert"),                    # Insert

    # emacs style
    C("Super-a"):               C("Home"),                      # Beginning of Line
    C("Super-e"):               C("End"),                       # End of Line
    C("Super-b"):               C("Left"),
    C("Super-f"):               C("Right"),
    C("Super-n"):               C("Down"),
    C("Super-p"):               C("Up"),
    C("Super-k"):              [C("Shift-End"), C("Backspace")],
    C("Super-d"):               C("Delete"),

    # C("Alt-RC-Space"):          C(""),                          # Open Finder - Placeholder not-deepin
    # C("Alt-RC-Space"):          C("Super-e"),                   # Open Finder - Placeholder (deepin)

    # Wordwise
    C("RC-Left"):               C("Home"),                      # Beginning of Line
    C("Shift-RC-Left"):         C("Shift-Home"),                # Select all to Beginning of Line
    C("RC-Right"):              C("End"),                       # End of Line
    C("Shift-RC-Right"):        C("Shift-End"),                 # Select all to End of Line
    # C("RC-Left"):               C("C-LEFT_BRACE"),              # Firefox-nw - Back
    # C("RC-Right"):              C("C-RIGHT_BRACE"),             # Firefox-nw - Forward
    # C("RC-Left"):               C("Alt-LEFT"),                  # Chrome-nw - Back
    # C("RC-Right"):              C("Alt-RIGHT"),                 # Chrome-nw - Forward
    C("RC-Up"):                 C("C-Home"),                    # Beginning of File
    C("Shift-RC-Up"):           C("C-Shift-Home"),              # Select all to Beginning of File
    C("RC-Down"):               C("C-End"),                     # End of File
    C("Shift-RC-Down"):         C("C-Shift-End"),               # Select all to End of File
    # C("RAlt-Backspace"):        C("Delete"),                    # Chromebook/IBM - Delete
    C("Super-Backspace"):       C("C-Backspace"),               # Delete Left Word of Cursor
    C("Super-Delete"):          C("C-Delete"),                  # Delete Right Word of Cursor
    # C("LAlt-Backspace"):        C("C-Backspace"),               # Chromebook/IBM - Delete Left Word of Cursor
    # C("Alt-Backspace"):         C("C-Backspace"),               # Default not-chromebook
    C("RC-Backspace"):          C("C-Shift-Backspace"),         # Delete Entire Line Left of Cursor
    C("Alt-Delete"):            C("C-Delete"),                  # Delete Right Word of Cursor
    # C(""):                      ignore_key,                     # cancel
    # C(""):                      C(""),                          #

# }, when = lambda ctx: ctx.wm_class.casefold() not in remotes) # original conditional
# }, when = matchProps(not_cls=remoteStr))                      # matchProps with regex string
}, when = matchProps(not_lst=remotes_lod))                      # matchProps with list-of-dicts
